#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""""""

__all__ = ["which_python_executable"]


from taurus.core.util.whichexecutable import whichfile


def which_python_executable():
    """Return full path to python executable.

    On some OS Python 3 is executed with python3 but in conda environments it
    is executed with python. Return Python 3 executable regardless of the
    Python installation.
    """
    executable = whichfile("python3")
    if executable is None:
        executable = whichfile("python")
    return executable



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains tests for trigger gate generation using a
given controller"""

import threading

from taurus.test import insertTest
import unittest

from sardana.pool.poolsynchronization import PoolSynchronization
from sardana.pool.poolacquisition import get_acq_ctrls
from sardana.sardanadefs import State
from sardana.pool.pooldefs import SynchDomain, SynchParam
from sardana.pool.test import FakePool, createCtrlConf, createElemConf, \
    createPoolController, createPoolTriggerGate, \
    createControllerConfiguration

__docformat__ = "restructuredtext"


class SynchronizationTestCase(object):
    """Base class for integration tests of PoolSynchronization class and any
    PoolTriggerGateController. Test is parameterized using trigger parameters.

    .. seealso:: :meth:`taurus.test.base.insertTest`"""

    def createElements(self, ctrl_klass, ctrl_lib, ctrl_props):
        # create controller and element
        ctrl_conf = createCtrlConf(self.pool, 'tgctrl01', ctrl_klass,
                                   ctrl_lib, ctrl_props)
        elem_conf = createElemConf(self.pool, 1, 'tg01')
        self.tg_ctrl = createPoolController(self.pool, ctrl_conf)
        self.tg_elem = createPoolTriggerGate(self.pool, self.tg_ctrl,
                                             elem_conf)
        # add controller and elements to containers
        self.tg_ctrl.add_element(self.tg_elem)
        self.pool.add_element(self.tg_ctrl)
        self.pool.add_element(self.tg_elem)
        # create Synchronization action and its configuration
        self.conf_ctrl = createControllerConfiguration(self.tg_ctrl,
                                                       [self.tg_elem])
        self.ctrls = get_acq_ctrls([self.conf_ctrl])
        self.tgaction = PoolSynchronization(self.tg_elem)
        self.tgaction.add_element(self.tg_elem)

    def setUp(self):
        """Create a FakePool object.
        """
        self.pool = FakePool()

    def tggeneration(self, ctrl_lib, ctrl_klass, ctrl_props,
                     synchronization):
        """Helper method to verify trigger element states before and after
        trigger/gate generation.

       :param ctrl_lib: controller library used for the test
       :type ctrl_lib: :obj:`str`
       :param ctrl_klass: controller class used for the test
       :type ctrl_klass: :obj:`str`
       :param offset: temporal offset before beginning the trigger generation
       :type offset: float
       :param active_interval: signal at which triggers will be generated
       :type active_interval: float
       :param passive_interval: temporal passive period between two active
                                periods
       :type passive_interval: float
       :param repetitions: number of generated triggers
       :type repetitions: int
        """

        # create controller and trigger element
        self.createElements(ctrl_klass, ctrl_lib, ctrl_props)

        # create start_action arguments
        args = (self.ctrls, synchronization)
        kwargs = {}
        # starting action
        self.tgaction.start_action(*args, **kwargs)
        # verifying that the elements involved in action changed its state
        element_state = self.tg_elem.get_state()
        msg = ("State after start_action is '%s'. (Expected: '%s')" %
               (State.get(element_state), "Moving"))
        self.assertEqual(element_state, State.Moving, msg)
        # entering action loop
        self.tgaction.action_loop()
        # verifying that the elements involved in action changed its state
        element_state = self.tg_elem.get_state()
        msg = ("State after action_loop shall be different than Moving")
        self.assertNotEqual(element_state, State.Moving, msg)

    def stopGeneration(self):
        """Method used to change the controller (mock) state"""
        self.tgaction.stop_action()

    def abort_tggeneration(self, ctrl_lib, ctrl_klass, ctrl_props,
                           synchronization, abort_time):
        """Helper method to verify trigger element states before and after
        trigger/gate generation when aborting the trigger generation.

       :param ctrl_lib: controller library used for the test
       :type ctrl_lib: :obj:`str`
       :param ctrl_klass: controller class used for the test
       :type ctrl_klass: :obj:`str`
       :param offset: temporal offset before beginning the trigger generation
       :type offset: float
       :param active_interval: signal at which triggers will be generated
       :type active_interval: float
       :param passive_interval: temporal passive period between two active
                                periods
       :type passive_interval: float
       :param repetitions: number of generated triggers
       :type repetitions: int
       :param abort_time: wait this time before stopping the trigger generation
       :type abort_time: float
        """

        # create controller and trigger element
        self.createElements(ctrl_klass, ctrl_lib, ctrl_props)

        # create start_action arguments
        args = (self.ctrls, synchronization)
        kwargs = {}
        # starting action
        self.tgaction.start_action(*args, **kwargs)
        # verifying that the elements involved in action changed its state
        element_state = self.tg_elem.get_state()
        msg = ("State after start_action is '%s'. (Expected: '%s')" %
               (State.get(element_state), "Moving"))
        self.assertEqual(element_state, State.Moving, msg)

        # starting timer (abort_time) stop the trigger generation
        threading.Timer(abort_time, self.stopGeneration).start()

        # entering action loop
        self.tgaction.action_loop()
        # verifying that the elements involved in action changed its state
        element_state = self.tg_elem.get_state()
        msg = ("State after action_loop shall be different than Moving")
        self.assertNotEqual(element_state, State.Moving, msg)

    def tearDown(self):
        self.tgaction = None
        self.tg_ctrl = None
        self.tg_cfg = None
        self.tg_elem = None


synchronization1 = [{SynchParam.Delay: {SynchDomain.Time: 0},
                     SynchParam.Active: {SynchDomain.Time: .01},
                     SynchParam.Total: {SynchDomain.Time: .02},
                     SynchParam.Repeats: 0}]

synchronization2 = [{SynchParam.Delay: {SynchDomain.Time: 0},
                     SynchParam.Active: {SynchDomain.Time: .01},
                     SynchParam.Total: {SynchDomain.Time: .02},
                     SynchParam.Repeats: 100}]


@insertTest(helper_name='tggeneration',
            ctrl_lib='DummyTriggerGateController',
            ctrl_klass='DummyTriggerGateController',
            ctrl_props={},
            synchronization=synchronization1
            )
@insertTest(helper_name='abort_tggeneration',
            ctrl_lib='DummyTriggerGateController',
            ctrl_klass='DummyTriggerGateController',
            ctrl_props={},
            synchronization=synchronization2,
            abort_time=0.5
            )
class DummySynchronizationTestCase(SynchronizationTestCase, unittest.TestCase):
    """Integration TestCase of Synchronization with \
       DummyTriggerGateController"""

    def setUp(self):
        unittest.TestCase.setUp(self)
        SynchronizationTestCase.setUp(self)

    def tearDown(self):
        SynchronizationTestCase.tearDown(self)
        unittest.TestCase.tearDown(self)



import os
import sys
import wx
import copy
import taurus
import settings_widget
import simple_tree_model
import wiz
import taurus.qt.qtgui.resource
from PyQt4 import QtGui, QtCore, Qt
from types import *
from taurus.core.util.enumeration import Enumeration
from sardana.taurus.core.tango.sardana import SardanaManager
from sardana.taurus.core.tango.sardana import PoolElementType

"""This wizard is created for adding new devices. At the moment it has 4 pages:

- SelectSardanaPool
- NewDevice
- CommitPage
- OutroPage
"""


class SelectSardanaPoolBasePage(wiz.SardanaBasePage):
    """Page for selecting Sardana and Pool

    Selected values are saved in

    self['sardana']
    self['pool']

    """

    def __init__(self, sardana=None, pool=None, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)
        self._sardana = sardana
        self._pool = pool

        self.setSubTitle('Please select the Pool from existing Sardana')
        self._valid = True
        self._panel = self.getPanelWidget()
        layout = QtGui.QGridLayout()
        self._sardanaNameCB = QtGui.QComboBox()
        self._poolNameCB = QtGui.QComboBox()
        layout.addItem(QtGui.QSpacerItem(
            60, 60, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum), 0, 0)
        layout.addWidget(QtGui.QLabel("Sardana"), 0, 1)
        layout.addItem(QtGui.QSpacerItem(
            60, 60, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum), 0, 3)
        layout.addWidget(self._sardanaNameCB, 0, 2)
        layout.addWidget(QtGui.QLabel("Pool"), 1, 1)
        layout.addWidget(self._poolNameCB, 1, 2)
        layout.addItem(QtGui.QSpacerItem(
            200, 10, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum), 2, 2)
        self._panel.setLayout(layout)
        self.setStatus(
            "Select the instances of Sardana and Pool, and click Next to continue")

        self['sardana'] = self._getSardana
        self['pool'] = self._getPool
        self.connect(self._sardanaNameCB, QtCore.SIGNAL(
            'currentIndexChanged(int)'), self._fillPoolNameCB)

    def initializePage(self):
        wiz.SardanaBasePage.initializePage(self)
        self._isSardanaProper = False
        if (self._sardana) and (self._sardana in self._getSardanas()):
            self._sardanaNameCB.addItem(self._sardana)
            self._sardanaNameCB.setEnabled(False)
            self._isSardanaProper = True
        else:
            self._fillSardanaNameCB()

        self._isPoolProper = False
        if (self._pool) and (self._isSardanaProper):
            for pool in self._getSardana().get_pools():
                if self._pool == pool.get_name():
                    self._isPoolProper = True

        if self._isPoolProper:
            self._poolNameCB.addItem(self._pool)
            self._poolNameCB.setEnabled(False)
        else:
            self._fillPoolNameCB()

    def isComplete(self):
        return self._valid

    def setNextPageId(self, id):
        self._nextPageId = id

    def nextId(self):
        return self._nextPageId

    def _fillSardanaNameCB(self):
        self._sardanaNameCB.clear()
        for item in self._getSardanas():
            self._sardanaNameCB.addItem(item)

    def _fillPoolNameCB(self, int=None):
        sardana = self._getSardana()
        self._poolNameCB.clear()
        for item in sardana.get_pools():
            self._poolNameCB.addItem(item.get_name())

    def _getSardanas(self):
        return SardanaManager().get_sardanas()

    def _getSardana(self):
        return self._getSardanas()[str(self._sardanaNameCB.currentText())]

    def _getPool(self):
        for pool in self._getSardana().get_pools():
            if str(self._poolNameCB.currentText()) == pool.get_name():
                return pool
        return None


class SimpleTreeView(QtGui.QTreeView):
    """
        extended version of QTreeView that is emitting signals of activation and expansion
    """

    def __init__(self, parent=None):
        super(SimpleTreeView, self).__init__(parent)
        self.setSelectionBehavior(QtGui.QTreeView.SelectItems)
        self.setUniformRowHeights(True)
#
        self.connect(self, QtCore.SIGNAL("activated(QModelIndex)"),
                     self.activated)
        self.connect(self, QtCore.SIGNAL("expanded(QModelIndex)"),
                     self.expanded)
        self.expanded()

    def currentFields(self):
        return self.model().asData(self.currentIndex())

    def activated(self, index):
        self.emit(QtCore.SIGNAL("activated"), self.model().asRecord(index))

    def expanded(self):
        if not self.model() is None:
            for column in range(self.model().columnCount(
                                QtCore.QModelIndex())):
                self.resizeColumnToContents(column)


class AxisSteper(Qt.QSpinBox):
    """ Object of this class is a Spin box that can jumps from 1 to the total number
        of axis while skipping busy values
    """

    def __init__(self, controllerInfo=None, busyValues=[], parent=None):
        Qt.QSpinBox.__init__(self, parent)
        self._busyValues = busyValues
        self._controllerInfo = controllerInfo
        self.setControllerInfo(controllerInfo)
        QtCore.QObject.connect(self.lineEdit(), QtCore.SIGNAL(
            "editingFinished()"), self._textEdited)

    def _textEdited(self):
        axis = int(self.value())
        if not self._controllerInfo.is_axis_free(axis):
            self._findFreeAxis()

    def getValue(self):
        return self.value()

    def setControllerInfo(self, controllerInfo):
        self._controllerInfo = controllerInfo
        if self._controllerInfo:
            self.setMaximum(self._controllerInfo.get_max_elements() - 1)
            self._findFreeAxis()

    def _findFreeAxis(self):
        for axis in range(self._controllerInfo.get_max_elements()):
            if self.is_axis_free(axis):
                self.setValue(axis)
                break

    def is_axis_free(self, axis):
        if self._controllerInfo.is_axis_free(axis):
            for busyAxis in self._busyValues:
                if axis == busyAxis:
                    return False
            return True
        else:
            return False

    def stepBy(self, step):
        i = 1
        while True:
            axis = self.value() + i * step
            if self.is_axis_free(self.value() + i * step):
                break
            i += 1
        if axis >= 0 and axis < self._controllerInfo.get_max_elements():
            self.setValue(axis)


class SingleAxisWidget(Qt.QWidget):
    """
        Widget for selecting name and axis for the device
    """

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        self._layout = QtGui.QGridLayout(self)
        self.setLayout(self._layout)
        self.setupUi()
        self._controllerInfo = None
        self._edited = False

    def setupUi(self):
        self._nameLabel = Qt.QLabel("Name:")
        self._nameLineEdit = Qt.QLineEdit()
        self._nameLineEdit.setMinimumSize(200, 25)
        self._layout.addWidget(self._nameLabel, 0, 0, 1, 1)
        self._layout.addWidget(self._nameLineEdit, 0, 1, 1, 1)
        self._monitorLabel = Qt.QLabel("Monitor List:")
        self._axisLabel = Qt.QLabel("Axis:")
        self._axisSpinBox = AxisSteper()
        self._layout.addWidget(self._axisLabel, 1, 0, 1, 1)
        self._layout.addWidget(self._axisSpinBox, 1, 1, 1, 1)
        QtCore.QObject.connect(self._axisSpinBox.lineEdit(), QtCore.SIGNAL(
            "textEdited(QString)"), self._textEdited)
        QtCore.QObject.connect(self._nameLineEdit, QtCore.SIGNAL(
            "textEdited(QString)"), self._textEdited)

    def _textEdited(self, str):
        self._edited = True

    def setControllerInfo(self, controllerInfo):
        self._nameLineEdit.setText("")
        self._edited = False
        self._controllerInfo = controllerInfo
        self._axisSpinBox.setControllerInfo(self._controllerInfo)

    def isEdited(self):
        return self._edited

    def getValue(self):
        return None


class NameEditorWidget(QtGui.QLineEdit):
    """
        for editing the name, it has to check the name if it exist in the sardana .. but it does
    """

    def __init__(self, parent=None):
        QtGui.QLineEdit.__init__(self, parent)
        self.setValue(self.getDefaultValue())

    def setValue(self, value):
        if value is None:
            value = self.getDefaultValue()
        self.setText(str(value))
        self._actualValue = self.getValue()

    def getValue(self):
        return str(self.text())

    def textChanged(self, string):
        QtGui.QLineEdit.textChanged(self, string)

    def focusOutEvent(self, event):  # QFocusEvent
        QtGui.QLineEdit.focusOutEvent(self, event)
        self.valueChanged()

    def valueChanged(self):
        if not (self.getValue() == self._actualValue):
            self.emit(QtCore.SIGNAL("valueChanged"),
                      self._actualValue, self.getValue())
        self._actualValue = self.getValue()
        self.setValue(self.getValue())  # if value is not valid

    @classmethod
    def getDefaultValue(self):
        return ""


class TableAxisDelegate(QtGui.QItemDelegate):
    """
    Item delegate for Table Axis Widget
    """

    def __init__(self, parent=None):
        QtGui.QItemDelegate.__init__(self, parent)

    def setControllerInfo(self, controllerInfo):
        self._controllerInfo = controllerInfo

    def createEditor(self, parent, option, index):
        if index.column() == 0:
            editor = NameEditorWidget(parent)
            editor.installEventFilter(self)
        if index.column() == 1:
            editor = AxisSteper(self._controllerInfo, self.parent(
            ).parent().getAxisList(), parent=parent)
            editor.installEventFilter(self)

        return editor

    def setEditorData(self, editor, index):
        if index.column() == 0:
            value = index.model().data(index, QtCore.Qt.DisplayRole)
            editor.setValue(value.toString())
        if index.column() == 1:
            value = index.model().data(index, QtCore.Qt.DisplayRole)
            int, bool = value.toInt()
            editor.setValue(int)

    def setModelData(self, editor, model, index):
        self.emit(QtCore.SIGNAL("editorValueChanged"),
                  editor.getValue(), index.row(), index.column())
        value = editor.getValue()
        if index.column() == 0:  # name
            model.setData(index, QtCore.QVariant(value))
        if index.column() == 1:  # axis
            if editor.is_axis_free(int(value)):
                model.setData(index, QtCore.QVariant(value))

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class TableAxisWidget (QtGui.QWidget):
    """
    Table for selecting multiple axis
    """

    def __init__(self, value=None, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self._controllerInfo = None
        self._layout = QtGui.QHBoxLayout(self)
        self.setLayout(self._layout)
        self._tableView = QtGui.QTableView()
        self._tableView.mousePressEvent = self.mousePressEvent
        self._tableView.setSelectionMode(QtGui.QTableView.SingleSelection)
        self._layout.addWidget(self._tableView)
        self._verticalLayout = QtGui.QVBoxLayout()
        self._addRowButton = QtGui.QPushButton(self)
        self._addRowButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-add"))
        self._addRowButton.setText("Add Device       ")
        self._verticalLayout.addWidget(self._addRowButton)
        self._removeRowButton = QtGui.QPushButton(self)
        self._removeRowButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("list-remove"))
        self._removeRowButton.setText("Remove Device    ")
        self._verticalLayout.addWidget(self._removeRowButton)
        self._upButton = QtGui.QPushButton(self)
        self._upButton.setIcon(taurus.qt.qtgui.resource.getThemeIcon("go-up"))
        self._upButton.setText("Move Up   ")
        self._verticalLayout.addWidget(self._upButton)
        self._downButton = QtGui.QPushButton(self)
        self._downButton.setIcon(
            taurus.qt.qtgui.resource.getThemeIcon("go-down"))
        self._downButton.setText("Move Down")
        self._verticalLayout.addWidget(self._downButton)
        spacerItem = QtGui.QSpacerItem(
            20, 20, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self._verticalLayout.addItem(spacerItem)
        QtCore.QObject.connect(
            self._addRowButton, QtCore.SIGNAL("clicked()"), self._addRow)
        QtCore.QObject.connect(self._removeRowButton,
                               QtCore.SIGNAL("clicked()"), self._removeRow)
        QtCore.QObject.connect(
            self._upButton, QtCore.SIGNAL("clicked()"), self._moveUp)
        QtCore.QObject.connect(
            self._downButton, QtCore.SIGNAL("clicked()"), self._moveDown)
        self._layout.addLayout(self._verticalLayout)
        self._delegate = TableAxisDelegate(self._tableView)
        #QtCore.QObject.connect(self._delegate, QtCore.SIGNAL("editorValueChanged"), self._valueChanged)
        self._tableView.setItemDelegate(self._delegate)
        QtCore.QObject.connect(self._delegate, QtCore.SIGNAL(
            "editorValueChanged"), self._textEdited)
        # self._tableView.setItemDelegate(self._delegate)
        sizePolicy = QtGui.QSizePolicy(
            QtGui.QSizePolicy.MinimumExpanding, QtGui.QSizePolicy.MinimumExpanding)
        self._tableView.setSizePolicy(sizePolicy)
        self._tableView.horizontalHeader().setDefaultSectionSize(80)
        self._tableView.horizontalHeader().setVisible(True)
        self._tableView.horizontalHeader().setStretchLastSection(True)
        self._tableView.setMinimumSize(QtCore.QSize(150, 150))
        self._edited = False

    def _textEdited(self, str=None):
        self._edited = True

    def isEdited(self):
        return self._edited

    def getAxisList(self):
        list = []
        for row in self.getValue():
            list.append(row[1])
        return list

    def _findFreeAxis(self):
        for axis in range(self._controllerInfo.get_max_elements()):
            if self.is_axis_free(axis):
                return axis

        return None

    def is_axis_free(self, axis):
        if self._controllerInfo.is_axis_free(axis):
            for busyAxis in self.getAxisList():
                if axis == busyAxis:
                    return False
            return True
        else:
            return False

    def setControllerInfo(self, controllerInfo):
        self.setValue([])
        self._controllerInfo = controllerInfo
        self._delegate.setControllerInfo(self._controllerInfo)
        self._edited = False

    def _getSelectedIndex(self):
        if len(self._tableView.selectedIndexes()):
            row = self._tableView.selectedIndexes()[0].row()
            column = self._tableView.selectedIndexes()[0].column()
        else:
            row = None
            column = None
        return [row, column]

    def _addRow(self):
        value = self.getValue()  # stored table
        rows = len(value)

        if rows < self._controllerInfo.get_max_elements():
            rowIndex = self._getSelectedIndex()[0]
            if rowIndex is None:
                rowIndex = rows
            value.insert(rowIndex, ["", self._findFreeAxis()])
            self.setValue(value)
            self._textEdited()

        if self._findFreeAxis() is None:
            self._addRowButton.setEnabled(False)

    def _removeRow(self):
        rowIndex = self._getSelectedIndex()[0]
        if rowIndex is not None:
            msgBox = QtGui.QMessageBox()
            msgBox.setText("Confirmation              ")
            msgBox.setStandardButtons(
                QtGui.QMessageBox().No | QtGui.QMessageBox().Yes)
            msgBox.setInformativeText("Remove device {%i} ?" % (rowIndex + 1))
            msgBox.setIcon(QtGui.QMessageBox.Question)
            ret = msgBox.exec_()
        if ret == QtGui.QMessageBox().Yes:
            value = self.getValue()
            self.setValue(value[:rowIndex] + value[rowIndex + 1:])
            self._textEdited()
            self._addRowButton.setEnabled(True)

    def _moveUp(self):
        value = self.getValue()  # stored table
        rows = len(value)
        rowIndex = self._getSelectedIndex()[0]
        if (rows > 1) and (rowIndex is not None) and (rowIndex > 0):
            x = value.pop(rowIndex)
            value.insert(rowIndex - 1, x)
            self.setValue(value)
            self._tableView.setCurrentIndex(
                self._tableView.model().index(rowIndex - 1, 0))
            self._textEdited()

    def _moveDown(self):
        value = self.getValue()  # stored table
        rows = len(value)
        rowIndex = self._getSelectedIndex()[0]
        if (rows > 1) and (rowIndex is not None) and (rowIndex < rows - 1):
            x = value.pop(rowIndex)
            value.insert(rowIndex + 1, x)
            self.setValue(value)
            self._tableView.setCurrentIndex(
                self._tableView.model().index(rowIndex + 1, 0))
            self._textEdited()

    def setValue(self, value):
        rows = len(value)
        columns = 2
        self._model = QtGui.QStandardItemModel(rows, columns)
        for row in range(rows):
            for column in range(columns):
                index = self._model.index(row, column, QtCore.QModelIndex())
                self._model.setData(index, QtCore.QVariant(value[row][column]))
        self._model.setHeaderData(0, Qt.Qt.Horizontal, Qt.QVariant("Name"))
        self._model.setHeaderData(1, Qt.Qt.Horizontal, Qt.QVariant("Axis"))
        self._tableView.setModel(self._model)
        self._tableView.horizontalHeader().setVisible(True)
        self._tableView.horizontalHeader().setStretchLastSection(True)

    def getValue(self):
        rows = self._model.rowCount()
        result = []
        for row in range(rows):
            records = []
            index = self._model.index(row, 0)
            name = self._model.data(index).toString()
            index = self._model.index(row, 1)
            axis, bool = self._model.data(index).toInt()
            result.append([name, axis])

        return copy.deepcopy(result)


class MultipleAxisWidget(Qt.QWidget):
    """
    Wizard for selecting multiple axis
    Method getValue() returns two dimensional array
        where the first column is name and the second is axis
    """

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        self._layout = QtGui.QGridLayout(self)
        self.setLayout(self._layout)
        self.setupUi()
        self._controllerInfo = None
        self._edited = False

    def setupUi(self):
        self._tableAxisWidget = TableAxisWidget()
        self._layout.addWidget(self._tableAxisWidget, 0, 0, 1, 1)

    def setControllerInfo(self, controllerInfo):
        self._edited = False
        self._controllerInfo = controllerInfo
        self._tableAxisWidget.setControllerInfo(self._controllerInfo)

    def isEdited(self):
        return self._tableAxisWidget.isEdited()

    def getValue(self):
        return self._tableAxisWidget.getValue()


class HardwareSettings(Qt.QWidget):
    """
    Page for editing single or multiple axis devices
    """

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        self._layout = QtGui.QGridLayout(self)
        self.setLayout(self._layout)
        self.setupUi()
        self._controllerInfo = None
        self._edited = False
        self._changeView()

    def setupUi(self):
        self.setVisible(False)
        self._singleButton = QtGui.QRadioButton(self)
        self._singleButton.setText("Single")
        self._singleButton.setChecked(True)
        self._layout.addWidget(self._singleButton, 0, 0)
        self._multipleButton = QtGui.QRadioButton(self)
        self._multipleButton.setText("Multiple")
        self._layout.addWidget(self._multipleButton, 0, 1)
        self._singleAxisWidget = SingleAxisWidget()
        self._layout.addWidget(self._singleAxisWidget, 1, 0, 1, 2)
        self._multipleAxisWidget = MultipleAxisWidget()
        self._layout.addWidget(self._multipleAxisWidget, 2, 0, 1, 2)

        QtCore.QObject.connect(self._singleButton, QtCore.SIGNAL(
            "clicked()"), self._changeView)
        QtCore.QObject.connect(self._multipleButton,
                               QtCore.SIGNAL("clicked()"), self._changeView)
        self._spacerItem1 = Qt.QSpacerItem(
            10, 0, Qt.QSizePolicy.Fixed, Qt.QSizePolicy.Expanding)
        self._layout.addItem(self._spacerItem1, 4, 0, 1, 1, Qt.Qt.AlignCenter)

    def _changeView(self):
        if self._singleButton.isChecked():
            self._singleAxisWidget.setVisible(True)
            self._multipleAxisWidget.setVisible(False)
        else:
            self._singleAxisWidget.setVisible(False)
            self._multipleAxisWidget.setVisible(True)

    def setControllerInfo(self, controllerInfo):
        self._controllerInfo = controllerInfo
        self._singleAxisWidget.setControllerInfo(self._controllerInfo)
        self._multipleAxisWidget.setControllerInfo(self._controllerInfo)

    def isEdited(self):
        return self._singleAxisWidget.isEdited() or self._multipleAxisWidget.isEdited()


class NewDeviceBasePage(wiz.SardanaBasePage):
    """
    Base Page for selecting and editing Devices
   """

    def __init__(self, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)
        self._valid = False
        QtGui.QWizardPage.__init__(self, parent)
        self._layout = QtGui.QGridLayout(self)
        self.setupUi()
        self.setLayout(self._layout)
        self.setTitle('Selecting device')
        self.connect(self._treeView, QtCore.SIGNAL("activated"),
                     self.activated)
        self._currentItem = None
        self._currentItemIndex = None

    def getProperties(self):
        if self._settings is not None:
            return self._settings.getProperties()
        else:
            return None

    def checkData(self):
        if self.picked():
            properties, name, values = self.getProperties()
            if len(name):
                self.setStatus("Editing: " + self.picked().get_name())
                self._valid = True
            else:
                self.setStatus('The name is invalid')
                self._valid = False
        else:
            self.setStatus('Please select a device')
            self._valid = False

        self.emit(QtCore.SIGNAL('completeChanged()'))

    def keyPressEvent(self, event):
        wiz.SardanaBasePage.keyPressEvent(self, event)
        if self._tabWidget.currentIndex() == 0:
            undo = (event.modifiers() ==
                    QtCore.Qt.ControlModifier and event.key() == QtCore.Qt.Key_Z)
            redo = (event.modifiers() ==
                    QtCore.Qt.ControlModifier and event.key() == QtCore.Qt.Key_Y)
            if (undo):
                self._settings.undo()
            if (redo):
                self._settings.redo()

    def initializePage(self):
        wiz.SardanaIntroBasePage.initializePage(self)
        self.wizard().__setitem__("properties", self.getProperties)
        # resizing application
        self._previousPageSize = copy.deepcopy(
            [self.wizard().size().width(), self.wizard().size().height()])
        preferedSize = [800, 600]
        desktopRect = QtGui.QApplication.desktop().availableGeometry(self)
        center = desktopRect.center()
        if (not self.wizard().isMaximized()) and (self.wizard().size().height() < preferedSize[1]) and (self.wizard().size().width() < preferedSize[0]):
            self.wizard().resize(preferedSize[0], preferedSize[1])
            self.wizard().move(center.x() - self.wizard().width() *
                               0.5, center.y() - self.wizard().height() * 0.5)
        self._pool = self.wizard()["pool"]
        self._loadTreeModel()
        self.checkData()

    def cleanupPage(self):
        wiz.SardanaIntroBasePage.cleanupPage(self)
        # setUp size for previous page
        preferedSize = copy.deepcopy(self._previousPageSize)
        desktopRect = QtGui.QApplication.desktop().availableGeometry(self)
        center = desktopRect.center()

        if (not self.wizard().isMaximized()) and (self.wizard().size().height() > preferedSize[1]) and (self.wizard().size().width() > preferedSize[0]):
            self.wizard().resize(preferedSize[0], preferedSize[1])
            self.wizard().move(center.x() - self.wizard().width() *
                               0.5, center.y() - self.wizard().height() * 0.5)

    def _loadTreeModel(self):

        self.model = simple_tree_model.SimpleTreeModel(self)
        controllerinfos = self._pool.get_controller_class_infos()
        controllers = self._pool.get_controller_infos()
        self.model.sort(-1)

        for item in controllerinfos:
            self.model.addRecord(["Controller", PoolElementType[
                                 item.get_controller_type()], item.get_name()], item, False)

        elementTypeList = sorted(PoolElementType.keys())
        for type in elementTypeList:
            self.model.addNodes([type], False)
        for item in controllers:
            self.model.addRecord(
                [PoolElementType[item.get_controller_type()], item.get_name()], item, False)

        self._treeView.setModel(self.model)

    def setupUi(self):
        self._treeView = SimpleTreeView(self)
        self._treeView.setSizePolicy(
            QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        self._layout.addWidget(self._treeView, 0, 0, 1, 1)
        self._tabWidget = QtGui.QTabWidget(self)
        self._settingsTab = QtGui.QWidget(self)
        self._horizontalLayout_2 = QtGui.QHBoxLayout(self._settingsTab)
        self._scrollArea = QtGui.QScrollArea(self._settingsTab)
        self._scrollArea.setFrameShape(QtGui.QFrame.NoFrame)
        self._scrollArea.setWidgetResizable(True)
        self._scrollAreaWidgetContents = QtGui.QWidget(self._scrollArea)
        self._scrollAreaWidgetContents.setGeometry(
            QtCore.QRect(0, 0, 263, 316))
        self._gridLayout_2 = QtGui.QGridLayout(self._scrollAreaWidgetContents)
        self._settings = settings_widget.SettingsWidget()
        self._hardwareSettings = HardwareSettings()
        self._gridLayout_2.addWidget(self._settings)
        self._scrollArea.setWidget(self._scrollAreaWidgetContents)
        self._horizontalLayout_2.addWidget(self._scrollArea)
        self._tabWidget.addTab(self._settingsTab, "Settings")
        self._descriptionTab = QtGui.QWidget()
        self._horizontalLayout = QtGui.QHBoxLayout(self._descriptionTab)
        self._scrollArea_2 = QtGui.QScrollArea(self._descriptionTab)
        self._scrollArea_2.setFrameShape(QtGui.QFrame.NoFrame)
        self._scrollArea_2.setWidgetResizable(True)
        self._scrollAreaWidgetContents_2 = QtGui.QWidget(self._scrollArea_2)
        self._gridLayout_3 = QtGui.QGridLayout(
            self._scrollAreaWidgetContents_2)
        self._description = DescriptionWidget()
        self._gridLayout_3.addWidget(self._description)
        self._scrollArea_2.setWidget(self._scrollAreaWidgetContents_2)
        self._horizontalLayout.addWidget(self._scrollArea_2)
        self._tabWidget.addTab(self._descriptionTab, "Description")
        self._tabWidget.setSizePolicy(
            QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)

        self._layout.addWidget(self._tabWidget, 0, 1, 1, 2)
        self._tabWidget.setCurrentIndex(0)
        self._status_label = QtGui.QLabel()
        self._status_label.setAutoFillBackground(True)
        palette = self._status_label.palette()
        gradient = QtGui.QLinearGradient(0, 0, 0, 15)
        gradient.setColorAt(0.0, Qt.QColor.fromRgb(60, 150, 255))
        gradient.setColorAt(0.5, Qt.QColor.fromRgb(0, 85, 227))
        gradient.setColorAt(1.0, Qt.QColor.fromRgb(60, 150, 255))
        gradient.setSpread(QtGui.QGradient.RepeatSpread)
        palette.setBrush(QtGui.QPalette.Window, QtGui.QBrush(gradient))
        palette.setBrush(QtGui.QPalette.WindowText, Qt.Qt.white)
        self._layout.addWidget(self._status_label, 1, 0, 1, 3)
        self._description.setVisible(False)
        self._settings.setVisible(False)

    def setStatus(self, text):
        self._status_label.setText(text)

    def picked(self):
        return self._treeView.currentFields()

    def activated(self, name):

        if (self.picked() is not None) and (self.picked() != self._currentItem):
            ret = QtGui.QMessageBox().Yes
            if self._settings.isEdited() or self._hardwareSettings.isEdited():
                msgBox = QtGui.QMessageBox()
                msgBox.setText("Confirmation              ")
                msgBox.setStandardButtons(
                    QtGui.QMessageBox().No | QtGui.QMessageBox().Yes)
                msgBox.setInformativeText(
                    "Do you want to leave the editor of %s without saving the changes?" % self._currentItem.get_name())
                msgBox.setIcon(QtGui.QMessageBox.Question)
                ret = msgBox.exec_()

            if ret == QtGui.QMessageBox().Yes:
                self._settings.setEdited(False)
                self._hardwareSettings.setControllerInfo(None)
                self._currentItem = self.picked()
                self._currentItemIndex = self._treeView.currentIndex()

                if isinstance(self._currentItem, taurus.core.tango.sardana.sardana.ControllerClassInfo):

                    self._gridLayout_2.addWidget(self._settings)
                    self._settings.setVisible(True)
                    self._gridLayout_2.removeWidget(self._hardwareSettings)
                    self._hardwareSettings.setVisible(False)
                    self._description.setVisible(True)
                    self._tabWidget.setTabEnabled(1, True)
                    self.setStatus("Editing: " + self.picked().get_name())

                    self._settings.setProperties(
                        self.picked().get_properties())
                    QtCore.QObject.connect(self._settings, QtCore.SIGNAL(
                        "propertyValueChanged()"), self.checkData)
                    self._description.setOrganization(
                        self.picked().get_organization())
                    self._description.setFamily(self.picked().get_family())
                    self._description.setModel(self.picked().get_model())
                    self._description.setDescription(
                        self.picked().get_description())
                    self._description.setImage(None)

                if isinstance(self._currentItem, taurus.core.tango.sardana.sardana.ControllerInfo):

                    self._gridLayout_2.removeWidget(self._settings)
                    self._settings.setVisible(False)
                    self._gridLayout_2.addWidget(self._hardwareSettings)
                    self._hardwareSettings.setVisible(True)
                    self._hardwareSettings.setControllerInfo(self._currentItem)
                    self._tabWidget.setTabEnabled(1, False)
                    self.setStatus("Editing: " + self.picked().get_name())
            else:
                self._treeView.selectionModel().setCurrentIndex(
                    self._currentItemIndex, QtGui.QItemSelectionModel.SelectCurrent)

        self.checkData()

    def setNextPageId(self, id):
        self._nextPageId = id

    def isComplete(self):
        return self._valid


class NewDeviceCommitBasePage(wiz.SardanaIntroBasePage):
    """
    The last commiting page on the wizard, doesn't work at the moment
    """

    def __init__(self, parent=None):
        QtGui.QWizardPage.__init__(self, parent)
        self.setPixmap(QtGui.QWizard.WatermarkPixmap,
                       QtGui.QPixmap(":/watermark.jpg"))
        self._layout = QtGui.QFormLayout()
        self.setCommitPage(True)
        self.setTitle('Confirmation')

    def next(self):
        QWizard.next(self)

    def _set_style(self, w):
        f = w.font()
        f.setBold(True)
        w.setFont(f)
        return w

    def initializePage(self):
        wiz.SardanaIntroBasePage.initializePage(self)
        self._previousPageSize = copy.deepcopy(
            [self.wizard().size().width(), self.wizard().size().height()])
        self.setPixmap(QtGui.QWizard.WatermarkPixmap,
                       QtGui.QPixmap(":/watermark.jpg"))
        self._spacerItem1 = QtGui.QSpacerItem(
            800, 600, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        self._layout.addItem(self._spacerItem1)
        preferedSize = [600, 600]  # prefered size for this page
        desktopRect = QtGui.QApplication.desktop().availableGeometry(self)
        center = desktopRect.center()
        if (not self.wizard().isMaximized()):
            self.wizard().resize(preferedSize[0], preferedSize[1])
            self.wizard().move(center.x() - self.wizard().width() *
                               0.5, center.y() - self.wizard().height() * 0.5)
            # and (self.wizard().size().height () > preferedSize[1]) and
            # (self.wizard().size().width() > preferedSize[0]):

    def cleanupPage(self):
        wiz.SardanaIntroBasePage.cleanupPage(self)
        self.setPixmap(QtGui.QWizard.WatermarkPixmap, QtGui.QPixmap())
        # setUp size for previous page
        preferedSize = copy.deepcopy(self._previousPageSize)
        desktopRect = QtGui.QApplication.desktop().availableGeometry(self)
        center = desktopRect.center()
        if (not self.wizard().isMaximized()) and (self.wizard().size().height() < preferedSize[1]) and (self.wizard().size().width() < preferedSize[0]):
            self.wizard().move(
                center.x() - preferedSize[0] * 0.5, center.y() - preferedSize[1] * 0.5)
            self.wizard().resize(preferedSize[0], preferedSize[1])

    def setNextPageId(self, id):
        self._nextPageId = id


class DescriptionWidget(QtGui.QWidget):
    """
    The part of the wizard that shows the object description (text + picture)
    """

    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.setupUi()

    def setupUi(self):
        self._imageSize = [100, 100]
        self._formLayout = QtGui.QFormLayout(self)
        self._organizationLabel = QtGui.QLabel("Organization:")
        self._formLayout.setWidget(
            0, QtGui.QFormLayout.LabelRole, self._organizationLabel)
        self._organizationLabelText = QtGui.QLabel(self)
        self._organizationLabelText.setFrameShape(QtGui.QFrame.Panel)
        self._organizationLabelText.setFrameShadow(QtGui.QFrame.Sunken)
        self._organizationLabelText.setWordWrap(False)
        self._formLayout.setWidget(
            0, QtGui.QFormLayout.FieldRole, self._organizationLabelText)
        self._familyLabel = QtGui.QLabel("Family:")
        self._formLayout.setWidget(
            1, QtGui.QFormLayout.LabelRole, self._familyLabel)
        self._familyLabelText = QtGui.QLabel(self)
        self._familyLabelText.setFrameShape(QtGui.QFrame.Panel)
        self._familyLabelText.setFrameShadow(QtGui.QFrame.Sunken)
        self._familyLabelText.setWordWrap(False)
        self._formLayout.setWidget(
            1, QtGui.QFormLayout.FieldRole, self._familyLabelText)
        self._modelLabel = QtGui.QLabel("Model:")
        self._formLayout.setWidget(
            2, QtGui.QFormLayout.LabelRole, self._modelLabel)
        self._modelLabelText = QtGui.QLabel(self)
        self._modelLabelText.setFrameShape(QtGui.QFrame.Panel)
        self._modelLabelText.setFrameShadow(QtGui.QFrame.Sunken)
        self._modelLabelText.setWordWrap(False)
        self._formLayout.setWidget(
            2, QtGui.QFormLayout.FieldRole, self._modelLabelText)
        self._descriptionLabel = QtGui.QLabel("Description:")
        self._formLayout.setWidget(
            3, QtGui.QFormLayout.LabelRole, self._descriptionLabel)
        self._descriptionLabelText = QtGui.QLabel(self)
        self._descriptionLabelText.setFrameShape(QtGui.QFrame.Panel)
        self._descriptionLabelText.setFrameShadow(QtGui.QFrame.Sunken)
        self._descriptionLabelText.setWordWrap(True)
        self._descriptionLabelText.setAlignment(QtCore.Qt.AlignTop)
        self._formLayout.setWidget(
            3, QtGui.QFormLayout.FieldRole, self._descriptionLabelText)
        self._imageLabel = QtGui.QLabel("Image:")
        self._formLayout.setWidget(
            4, QtGui.QFormLayout.LabelRole, self._imageLabel)
        self._deviceLogo = QtGui.QLabel(self)
        self._deviceLogo.setPixmap(taurus.qt.qtgui.resource.getThemePixmap(
            "image-missing").scaled(*self._imageSize))
        self._deviceLogo.pixmap()
        self._deviceLogo.setAlignment(QtCore.Qt.AlignHCenter)
        self._formLayout.setWidget(
            4, QtGui.QFormLayout.FieldRole, self._deviceLogo)

    def setOrganization(self, name):
        self._organizationLabelText.setText(name)

    def setFamily(self, name):
        self._familyLabelText.setText(name)

    def setModel(self, name):
        self._modelLabelText.setText(name)

    def setDescription(self, text):
        self._descriptionLabelText.setText(text)

    def setImage(self, image):
        if isinstance(image, QtGui.QPixmap):
            self._deviceLogo.setPixmap(image.scaled(*self._imageSize))
        elif isinstance(image, QtGui.QImage):
            self._deviceLogo.setPixmap(
                QtGui.QPixmap().fromImage(image).scaled(*self._imageSize))
        else:
            self._deviceLogo.setPixmap(
                taurus.qt.qtgui.resource.getThemePixmap("image-missing").scaled(50, 50))


def addDevice(Sardana=None, Pool=None):
    app = QtGui.QApplication([])
    QtCore.QResource.registerResource(get_resources())
    Pages = Enumeration('Pages', ('SelectSardanaPool',
                                  'NewDevice', 'CommitPage', 'OutroPage'))
    w = wiz.SardanaBaseWizard()
    w.setWindowTitle("Add New Hardware Wizard")
    selectPool = SelectSardanaPoolBasePage(Sardana, Pool)
    w.setPage(Pages.SelectSardanaPool, selectPool)
    selectPool.setNextPageId(Pages.NewDevice)
    newDevice = NewDeviceBasePage()
    w.setPage(Pages.NewDevice, newDevice)
    newDevice.setNextPageId(Pages.CommitPage)
    commit_page = NewDeviceCommitBasePage()
    w.setPage(Pages.CommitPage, commit_page)
    commit_page.setNextPageId(Pages.OutroPage)
    w.show()
    sys.exit(app.exec_())


def get_resources():
    res_fname = os.path.abspath(__file__)
    res_fname = os.path.splitext(res_fname)[0] + '.rcc'
    return res_fname

if __name__ == "__main__":
    addDevice()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Pool utils"""

__all__ = ["PoolUtil"]

__docformat__ = 'restructuredtext'

from taurus.core.util.containers import CaselessDict
import threading


class _PoolUtil(object):

    def __init__(self):
        self._ctrl_proxies = CaselessDict()
        self._lock = threading.Lock()

    def __call__(self, *args, **kwargs):
        return self

    def get_device(self, *args, **kwargs):
        """Factory method to create a single `tango.DeviceProxy` instance
        per controller instance.

        :param ctrl_name: Controller name to which assign the proxy object
        :type ctrl_name: `str`
        :param device_name: Tango device name
        :type device_name: `str`
        :return: single device proxy object
        :rtype: `tango.DeviceProxy`
        """
        ctrl_name = args[0]
        device_name = args[1]
        with self._lock:
            ctrl_devs = self._ctrl_proxies.get(ctrl_name)
            if ctrl_devs is None:
                self._ctrl_proxies[ctrl_name] = ctrl_devs = CaselessDict()
            dev = ctrl_devs.get(device_name)
            if dev is None:
                import PyTango
                ctrl_devs[device_name] = dev = PyTango.DeviceProxy(device_name)
        return dev

    get_motor = get_phy_motor = get_pseudo_motor = get_motor_group = \
        get_exp_channel = get_ct_channel = get_zerod_channel = \
        get_oned_channel = get_twod_channel = get_pseudo_counter_channel = \
        get_measurement_group = get_com_channel = get_ioregister = get_device


#: Singleton instance of the `~sardana.pool.poolutil._PoolUtil` class.
#:
#: It is a factory of `tango.DeviceProxy` objects and ensures only one
#: instance of such objects is created for the whole process.
#: Please refer to the `~sardana.pool.poolutil._PoolUtil` API on the available
#: methods.
PoolUtil = _PoolUtil()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import pytest

from sardana import State
from sardana.pool.test import (FakePool, createPoolController,
                               createPoolMotor, dummyPoolMotorCtrlConf01,
                               dummyMotorConf01)


def StateOne_state(self, axis):
    return State.On


def StateOne_state_status(self, axis):
    return State.On, "Status"


def StateOne_state_status_limits(self, axis):
    return State.On, "Status", 0


@pytest.mark.parametrize("mock_StateOne", [StateOne_state,
                                           StateOne_state_status,
                                           StateOne_state_status_limits])
def test_state(monkeypatch, mock_StateOne):
    """Test variants of StateOne return value:
    - state
    - state, status
    - state, status, limit_switches
    """
    pool = FakePool()
    # when SEP19 gets implemented it should be possible to mock directly
    # the imported class
    DummyMotorController = pool.ctrl_manager.getControllerClass(
        "DummyMotorController")
    monkeypatch.setattr(DummyMotorController, "StateOne", mock_StateOne)
    mot_ctrl = createPoolController(pool, dummyPoolMotorCtrlConf01)
    mot = createPoolMotor(pool, mot_ctrl, dummyMotorConf01)
    mot_ctrl.add_element(mot)
    pool.add_element(mot_ctrl)
    pool.add_element(mot)
    assert mot.state == State.On
    assert type(mot.status) == str
    assert mot.limit_switches.value == (False, ) * 3



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import unittest

from sardana.macroserver.macros.test import RunMacroTestCase, testRun
from sardana.tango.macroserver.test import BaseMacroServerTestCase
from sardana.tango.pool.test.test_measurementgroup import MeasSarTestTestCase
from sardana.macroserver.macros.test.test_scanct import mg_config4


@testRun(macro_name="lsgh", wait_timeout=1)
@testRun(macro_name="defgh", macro_params=["lsm", "pre-acq"], wait_timeout=1)
@testRun(macro_name="defgh", macro_params=["lsm mot.*", "pre-acq"],
         wait_timeout=1)
@testRun(macro_name="udefgh", wait_timeout=1)
class GeneralHooksMacrosTest(RunMacroTestCase, unittest.TestCase):
    pass


class GeneralHooksTest(MeasSarTestTestCase, BaseMacroServerTestCase,
                       RunMacroTestCase, unittest.TestCase):

    def setUp(self):
        MeasSarTestTestCase.setUp(self)
        BaseMacroServerTestCase.setUp(self)
        RunMacroTestCase.setUp(self)
        unittest.TestCase.setUp(self)

    def create_meas(self, config):
        MeasSarTestTestCase.create_meas(self, config)
        self.macro_executor.run(macro_name='senv',
                                macro_params=['ActiveMntGrp', '_test_mg_1'],
                                sync=True, timeout=1.)

    def test_gh(self):
        self.macro_runs(macro_name="defgh", macro_params=["lsm", "pre-acq"],
                        wait_timeout=1)
        self.create_meas(mg_config4)
        self.macro_runs(macro_name="ct", macro_params=[".1"], wait_timeout=1)
        self.macro_runs(macro_name="udefgh", macro_params=["lsm", "pre-acq"],
                        wait_timeout=1)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        RunMacroTestCase.tearDown(self)
        BaseMacroServerTestCase.tearDown(self)
        MeasSarTestTestCase.tearDown(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package provides the spock macroserver connectivity"""

__all__ = ['GUIViewer', 'SpockBaseDoor', 'QSpockDoor', 'SpockDoor',
           'SpockMacroServer']

import os
import ctypes
import PyTango

from taurus.core import TaurusEventType, TaurusSWDevState, TaurusDevState

from sardana.sardanautils import is_pure_str, is_non_str_seq
from sardana.spock import genutils
from sardana.util.parser import ParamParser
from sardana.spock.inputhandler import SpockInputHandler, InputHandler
from sardana import sardanacustomsettings

CHANGE_EVTS = TaurusEventType.Change, TaurusEventType.Periodic


if genutils.get_gui_mode() == 'qt':
    from sardana.taurus.qt.qtcore.tango.sardana.macroserver import QDoor, QMacroServer
    BaseDoor = QDoor
    BaseMacroServer = QMacroServer
    BaseGUIViewer = object
else:
    from sardana.taurus.core.tango.sardana.macroserver import BaseDoor, BaseMacroServer
    BaseGUIViewer = object


RUNNING_STATE = TaurusDevState.Ready


class GUIViewer(BaseGUIViewer):

    def __init__(self, door=None):
        BaseGUIViewer.__init__(self)
        self._door = door

    def run(self):
        self.plot()

    def show_scan(self, scan_nb=None, scan_history_info=None, directory_map=None):
        scan_dir, scan_file = None, None
        if scan_nb is None:
            import h5py
            for scan in reversed(scan_history_info):
                scan_dir = scan.get('ScanDir')
                scan_file = scan.get('ScanFile')
                if scan_dir is None or scan_file is None:
                    continue
                if not isinstance(scan_file, str):
                    scan_files = scan_file
                    scan_file = None
                    for fname in scan_files:
                        try:
                            h5py.File(os.path.join(scan_dir, fname), "r")
                        except IOError:
                            pass
                        else:
                            scan_file = fname
                            break
                    if scan_file is None:
                        print("Cannot plot scan:")
                        print("It only works with HDF5 files.")
                        return
                else:
                    try:
                        h5py.File(os.path.join(scan_dir, scan_file), "r")
                    except IOError:
                        print("Cannot plot scan:")
                        print("It only works with HDF5 files.")
                        return
                break
            else:
                print("Cannot plot scan:")
                print("No scan in scan history was saved into a file")
                return
        else:
            for scan in reversed(scan_history_info):
                if scan['serialno'] == scan_nb:
                    scan_dir = scan.get('ScanDir')
                    scan_file = scan.get('ScanFile')
                    if scan_dir is None or scan_file is None:
                        print("Cannot plot scan:")
                        print("Scan %d was not saved into a file" % (scan_nb,))
                        return
                    if not isinstance(scan_file, str):
                        scan_file = scan_file[0]
                    break
            else:
                print("Cannot plot scan:")
                print("Scan %d not found in scan history" % (scan_nb,))
                return

        remote_file = os.path.join(scan_dir, scan_file)

        locations = [scan_dir]
        local_file = None
        if directory_map is None or not scan_dir in directory_map:
            if os.path.isdir(scan_dir):
                if scan_file in os.listdir(scan_dir):
                    local_file = remote_file
        else:
            local_directories = directory_map[scan_dir]
            # local_directories may be either a string or a list of strings
            # the further logic is unified to work on a list, so convert it
            if isinstance(local_directories, str):
                local_directories = [local_directories]
            locations = local_directories
            if scan_dir not in locations:
                locations.append(scan_dir)
            for local_directory in local_directories:
                if os.path.isdir(local_directory):
                    if scan_file in os.listdir(local_directory):
                        local_file = os.path.join(local_directory, scan_file)
                        break
        if local_file is None:
            print("Cannot plot scan:")
            print("Could not find %s in any of the following locations:" %
                  (scan_file,))
            print("\n".join(locations))
            return

        import taurus.qt.qtgui.extra_nexus
        taurus_nexus_widget = taurus.qt.qtgui.extra_nexus.TaurusNeXusBrowser()
        taurus_nexus_widget.setMinimumSize(800, 600)

        print("Trying to open local scan file %s..." % (local_file,))
        taurus_nexus_widget.openFile(local_file)
        taurus_nexus_widget.show()
        nexus_widget = taurus_nexus_widget.neXusWidget()
        entry_name = "entry%d" % scan["serialno"]
        measurement_name = "%s/measurement" % entry_name
        title_name = "%s/title" % entry_name
        windowTitle = scan_file + "[" + entry_name + "]"

        try:
            #entry_index = taurus_nexus_widget.findNodeIndex(local_file, entry_name)
            measurement_index = taurus_nexus_widget.findNodeIndex(
                local_file, measurement_name)
            # nexus_widget.setRootIndex(entry_index)
            nexus_widget.setCurrentIndex(measurement_index)
            nexus_widget.expand(measurement_index)
            title_index = taurus_nexus_widget.findNodeIndex(
                local_file, title_name)
            file_model = nexus_widget.model()
            title = file_model.getNodeFromIndex(title_index)[0]
            windowTitle += " - " + title
        except Exception as e:
            print("Cannot plot scan:")
            print(str(e))

        taurus_nexus_widget.setWindowTitle(windowTitle)

    def plot(self):
        try:
            import sps
        except Exception:
            print('sps module not available. No plotting')
            return

        try:
            import pylab
        except Exception:
            print("pylab not available (try running 'spock -pylab'). "
                  "No plotting")
            return

        door = genutils.get_door()

        try:
            env = dict(door.getEnvironmentObj().read().value)
        except Exception as e:
            print('Unable to read environment. No plotting')
            print(str(e))
            return

        program = door.getNormalName().replace('/', '').replace('_', '')
        try:
            array = env['ActiveMntGrp'].replace(
                '/', '').replace('_', '').upper() + "0D"
            array_ENV = '%s_ENV' % array
        except:
            print('ActiveMntGrp not defined. No plotting')
            return

        if not program in sps.getspeclist():
            print('%s not found. No plotting' % program)
            return

        if not array in sps.getarraylist(program):
            print('%s not found in %s. No plotting' % (array, program))
            return

        if not array_ENV in sps.getarraylist(program):
            print('%s not found in %s. No plotting' % (array_ENV, program))
            return

        try:
            mem = sps.attach(program, array)
            mem_ENV = sps.attach(program, array_ENV)
        except Exception as e:
            print('sps.attach error: %s. No plotting' % str(e))
            return

        # reconstruct the environment
        i, env = 0, {}
        while mem_ENV[i][0] != '':
            line = mem_ENV[i].tostring()
            eq, end = line.index('='), line.index('\x00')
            k, v = line[:eq], line[eq + 1:end]
            env[k] = v
            i += 1

        labels = env['axistitles'].split(' ')

        col_nb = len(labels)

        if col_nb < 4:
            print('No data columns available in sps')
            return

        rows = int(env['nopts'])

        m = mem.transpose()

        x = m[1][:rows]
        colors = 'bgrcmyk'
        col_nb = min(col_nb, len(colors) + 3)
        # skip point_nb, motor and timer columns
        for i in range(3, col_nb):
            y = m[i][:rows]
            line, = pylab.plot(x, y, label=labels[i])
            line.linestyle = '-'
            line.linewidth = 1
            line.color = colors[i - 3]
        pylab.legend()


class SpockBaseDoor(BaseDoor):
    """A CLI version of the Door device"""

    console_editors = 'vi', 'vim', 'nano', 'joe', 'pico', 'emacs'

    Critical = 'Critical'
    Error = 'Error'
    Info = 'Info'
    Warning = 'Warning'
    Output = 'Output'
    Debug = 'Debug'
    Result = 'Result'
    RecordData = 'RecordData'

    MathFrontend = "matplotlib"

    def __init__(self, name, **kw):
        self._consoleReady = kw.get("consoleReady", False)
        if 'silent' not in kw:
            kw['silent'] = False
        self._lines = []
        self._spock_state = None
        self._plotter = GUIViewer(self)
        self.call__init__(BaseDoor, name, **kw)

    def create_input_handler(self):
        return SpockInputHandler(self)

    def get_color_mode(self):
        return genutils.get_color_mode()

    def _get_macroserver_for_door(self):
        ret = genutils.get_macro_server()
        return ret

    def _preprocessParameters(self, parameters):
        if is_pure_str(parameters):
            inside_str = False
            pars = []
            par = ''
            for c in parameters:
                if c == '"':
                    if inside_str:
                        inside_str = False
                        pars.append(par)
                        par = ''
                    else:
                        inside_str = True
                elif c == ' ':
                    if inside_str:
                        par += c
                    else:
                        pars.append(par)
                        par = ''
                else:
                    par += c
            if par:
                pars.append(par)
            return pars
        elif is_non_str_seq(parameters):
            return parameters

    def preRunMacro(self, obj, parameters):
        return BaseDoor.preRunMacro(self, obj, self._preprocessParameters(parameters))

    def runMacro(self, obj, parameters=[], synch=False):
        return BaseDoor.runMacro(self, obj, parameters=parameters, synch=synch)

    def _handle_second_release(self):
        try:
            self.write('4th Ctrl-C received: Releasing...\n')
            self.release()
            self.block_lines = 0
            self.release()
            self.writeln("Releasing done!")
        except KeyboardInterrupt:
            self.write('5th Ctrl-C received: Giving up...\n')
            self.write('(The macro is probably still executing the on_abort '
                       'method. Either wait or restart the server.)\n')

    def _handle_first_release(self):
        try:
            self.write('3rd Ctrl-C received: Releasing...\n')
            self.block_lines = 0
            self.release()
            self.writeln("Releasing done!")
        except KeyboardInterrupt:
            self._handle_second_release()

    def _handle_abort(self):
        try:
            self.write('2nd Ctrl-C received: Aborting...\n')
            self.block_lines = 0
            self.abort()
            self.writeln("Aborting done!")
        except KeyboardInterrupt:
            self._handle_first_release()

    def _handle_stop(self):
        try:
            self.write('\nCtrl-C received: Stopping...\n')
            self.block_lines = 0
            self.stop()
            self.writeln("Stopping done!")
        except KeyboardInterrupt:
            self._handle_abort()

    def _runMacro(self, xml, **kwargs):
        # kwargs like 'synch' are ignored in this re-implementation
        if self._spock_state != RUNNING_STATE:
            print("Unable to run macro: No connection to door '%s'" %
                  self.getSimpleName())
            raise Exception("Unable to run macro: No connection")
        if self.stateObj.read().rvalue == PyTango.DevState.RUNNING:
            print("Another macro is running. Wait until it finishes...")
            raise Exception("Unable to run macro: door in RUNNING state")
        if xml is None:
            xml = self.getRunningXML()
        kwargs['synch'] = True
        try:
            return BaseDoor._runMacro(self, xml, **kwargs)
        except KeyboardInterrupt:
            self._handle_stop()
        except PyTango.DevFailed as e:
            if is_non_str_seq(e.args) and \
               not isinstance(e.args, str):
                reason, desc = e.args[0].reason, e.args[0].desc
                macro_obj = self.getRunningMacro()
                if reason == 'MissingParam':
                    print("Missing parameter:", desc)
                    print(macro_obj.getInfo().doc)
                elif reason == 'WrongParam':
                    print("Wrong parameter:", desc)
                    print(macro_obj.getInfo().doc)
                elif reason == 'UnkownParamObj':
                    print("Unknown parameter:", desc)
                elif reason == 'MissingEnv':
                    print("Missing environment:", desc)
                elif reason in ('API_CantConnectToDevice',
                                'API_DeviceNotExported'):
                    self._updateState(self._old_sw_door_state,
                                      TaurusSWDevState.Shutdown, silent=True)
                    print("Unable to run macro: No connection to door '%s'" %
                          self.getSimpleName())
                else:
                    print("Unable to run macro:", reason, desc)

    def _getMacroResult(self, macro):
        ret = None
        if macro.info.hasResult():
            ret = macro.getResult()

            if ret is None:
                return None

            if macro.info.getResult().type == 'File':

                commit_cmd = macro.info.hints['commit_cmd']

                if commit_cmd is None:
                    return ret

                local_f_name = ret[0]
                remote_f_name = ret[1]
                line_nb = ret[3]
                commit = CommitFile(commit_cmd, local_f_name, remote_f_name)
                self.pending_commits.update({remote_f_name: commit})
                ip = genutils.get_ipapi()
                editor = genutils.get_editor()

                cmd = 'edit -x -n %s %s' % (line_nb, local_f_name)
                if editor not in self.console_editors:
                    cmd = 'bg _ip.magic("' + cmd + '")'
                ip.magic(cmd)
                # The return value of the macro was saved in a file and opened
                # with edit so we don't return anything to avoid big outputs
                # to the console
                ret = None
        return ret

    def plot(self):
        self._plotter.run()

    def show_scan(self, scan_nb=None):
        env = self.getEnvironment()
        scan_history_info = env.get("ScanHistory")
        directory_map = env.get("DirectoryMap")
        self._plotter.show_scan(scan_nb=scan_nb,
                                scan_history_info=scan_history_info,
                                directory_map=directory_map)

    def stateChanged(self, s, t, v):
        old_sw_state = self._old_sw_door_state
        BaseDoor.stateChanged(self, s, t, v)
        new_sw_state = self._old_sw_door_state
        self._updateState(old_sw_state, new_sw_state)

    def _updateState(self, old_sw_state, new_sw_state, silent=False):
        user_ns = genutils.get_ipapi().user_ns
        if new_sw_state == RUNNING_STATE:
            user_ns['DOOR_STATE'] = ""
        else:
            user_ns['DOOR_STATE'] = " (OFFLINE)"

        if not self.isConsoleReady():
            self._spock_state = new_sw_state
            return

        ss = self._spock_state
        if ss is not None and ss != new_sw_state and not silent:
            if ss == RUNNING_STATE:
                self.write_asynch(
                    "\nConnection to door '%s' was lost.\n" % self.getSimpleName())
            elif new_sw_state == RUNNING_STATE:
                self.write_asynch("\nConnection to the door (%s) has "
                                  "been restablished\n" % self.getSimpleName())
        self._spock_state = new_sw_state

    def write_asynch(self, msg):
        self._lines.append(msg)

    def pre_prompt_hook(self, ip):
        self._flush_lines()

    def _flush_lines(self):
        for l in self._lines:
            self.write(l)
        self._lines = []

    def setConsoleReady(self, state):
        self._consoleReady = state

    def isConsoleReady(self):
        return self._consoleReady

    def write(self, msg, stream=None):
        if not self.isConsoleReady():
            return
        return BaseDoor.write(self, msg, stream=stream)

    def processRecordData(self, data):
        if data is None:
            return
        data = data[1]
        if (isinstance(data, dict)
                and 'type' in data
                and data['type'] == 'function'):
            func_name = data['func_name']
            if func_name.startswith("pyplot."):
                func_name = self.MathFrontend + "." + func_name
            args = data['args']
            kwargs = data['kwargs']

            members = func_name.split(".")
            mod_list, fname = members[:-1], members[-1]
            mod_name = ".".join(mod_list)
            if mod_name:
                mod = __import__(mod_name, fromlist=mod_list)
                func = getattr(mod, fname)
            else:
                func = __builtins__[fname]
            try:
                func(*args, **kwargs)
            except Exception as e:
                self.logReceived(
                    self.Warning, ['Unable to execute %s: ' % func_name, str(e)])

    _RECORD_DATA_THRESOLD = 4 * 1024 * 1024  # 4Mb

    def _processInput(self, input_data):
        pyos_inputhook_ptr = ctypes.c_void_p.in_dll(
            ctypes.pythonapi, "PyOS_InputHook")
        old_pyos_inputhook_ptr = pyos_inputhook_ptr.value
        pyos_inputhook_ptr.value = ctypes.c_void_p(None).value
        ret = BaseDoor._processInput(self, input_data)
        pyos_inputhook_ptr.value = old_pyos_inputhook_ptr
        return ret

    def _processRecordData(self, data):
        if data is None:
            return
        value = data.value
        size = len(value[1])
        if size > self._RECORD_DATA_THRESOLD:
            sizekb = size // 1024
            self.logReceived(self.Info, ['Received long data record (%d Kb)' % sizekb,
                                         'It may take some time to process. Please wait...'])
        return BaseDoor._processRecordData(self, data)


from taurus.external.qt import Qt


class QSpockDoor(SpockBaseDoor):

    def __init__(self, name, **kw):
        self.call__init__(SpockBaseDoor, name, **kw)

        self.recordDataUpdated.connect(self.processRecordData)

    def recordDataReceived(self, s, t, v):
        if genutils.get_pylab_mode() == "inline":
            if t not in CHANGE_EVTS:
                return
            res = BaseDoor.recordDataReceived(self, s, t, v)
            self.processRecordData(res)
        else:
            res = SpockBaseDoor.recordDataReceived(self, s, t, v)
        return res

    def create_input_handler(self):
        inputhandler = getattr(sardanacustomsettings, 'SPOCK_INPUT_HANDLER',
                               "CLI")

        if inputhandler == "Qt":
            return InputHandler()
        else:
            return SpockInputHandler()


class SpockDoor(SpockBaseDoor):

    def _processRecordData(self, data):
        data = SpockBaseDoor._processRecordData(self, data)
        return self.processRecordData(data)


class SpockMacroServer(BaseMacroServer):
    """A CLI version of the MacroServer device"""

    def __init__(self, name, **kw):
        self._local_magic = {}
        self._local_var = set()
        self.call__init__(BaseMacroServer, name, **kw)

    def on_elements_changed(self, evt_src, evt_type, evt_value):
        return BaseMacroServer.on_elements_changed(self, evt_src, evt_type,
                                                   evt_value)

    _SKIP_ELEMENTS = 'controller', 'motorgroup', 'instrument', \
        'controllerclass', 'controllerlib', 'macrolib'

    def _addElement(self, element_data):
        element = BaseMacroServer._addElement(self, element_data)
        elem_type = element.type
        if "MacroCode" in element.interfaces:
            self._addMacro(element)
        elif elem_type not in self.NO_CLASS_TYPES:
            # TODO: when it becomes possible to do:
            # some taurus.Device.<attr name> = <value>
            # replace device_proxy with element
            device_proxy = element.getObj().getDeviceProxy()
            genutils.expose_variable(element.name, device_proxy)
        return element

    def _removeElement(self, element_data):
        element = BaseMacroServer._removeElement(self, element_data)
        elem_type = element.type
        if "MacroCode" in element.interfaces:
            self._removeMacro(element)
        elif elem_type not in self.NO_CLASS_TYPES:
            genutils.unexpose_variable(element.name)
        return element

    def _addMacro(self, macro_info):
        macro_name = str(macro_info.name)

        # IPython < 1 magic commands have different API
        if genutils.get_ipython_version_list() < [1, 0]:
            def macro_fn(shell, parameter_s='', name=macro_name):
                door = genutils.get_door()
                ms = genutils.get_macro_server()
                params_def = ms.getMacroInfoObj(name).parameters
                parameters = split_macro_parameters(parameter_s, params_def)
                door.runMacro(macro_name, parameters, synch=True)
                macro = door.getLastRunningMacro()
                if macro is not None:  # maybe none if macro was aborted
                    return macro.getResult()
        else:
            def macro_fn(parameter_s='', name=macro_name):
                door = genutils.get_door()
                ms = genutils.get_macro_server()
                params_def = ms.getMacroInfoObj(name).parameters
                parameters = split_macro_parameters(parameter_s, params_def)
                door.runMacro(macro_name, parameters, synch=True)
                macro = door.getLastRunningMacro()
                if macro is not None:  # maybe none if macro was aborted
                    return macro.getResult()

        macro_fn.__name__ = macro_name
        macro_fn.__doc__ = macro_info.doc + "\nWARNING: do not rely on the" \
                                            " file path below\n"

        # register magic command
        genutils.expose_magic(macro_name, macro_fn)
        self._local_magic[macro_name] = macro_fn

        return macro_info

    def _removeMacro(self, macro_info):
        macro_name = macro_info.name
        genutils.unexpose_magic(macro_name)
        del self._local_magic[macro_name]


def split_macro_parameters(parameters_s, params_def):
    """Split string with macro parameters into a list with macro parameters.
    Whitespaces are the separators between the parameters.

    Repeat parameters are encapsulated in square brackets and its internal
    repetitions, if composed from more than one item are also encapsulated
    in brackets. In this case the output list contains lists internally.

    :param parameters_s: input string containing parameters
    :type parameters_s: string
    :return:  parameters represented as a list (may contain internal lists
    :rtype: list
    """
    parser = ParamParser(params_def)
    return parser.parse(parameters_s)



import PyTango
import socket
import MotorController


class IcePapController(MotorController.MotorController):
    "This class is the Tango Sardana motor controller for the ICEPAP with properties"

    ctrl_features = ['Encoder', 'Home_speed', 'Home_acceleration']

    class_prop = {'A_Class_Prop': {'Type': 'PyTango.DevLong', 'Description': 'A class property', 'DefaultValue': 20},
                  'Another_Class_Prop': {'Type': 'PyTango.DevDouble', 'Description': 'Bla bla bla', 'DefaultValue': 2.345},
                  'Host': {'Type': 'PyTango.DevString', 'Description': 'The host name'},
                  'Port': {'Type': 'PyTango.DevLong', 'Description': 'The port number'}}

    ctrl_extra_attributes = {'First_extra': {'Type': 'PyTango.DevDouble', 'R/W Type': 'PyTango.READ_WRITE'},
                             'Second_extra': {'Type': 'PyTango.DevLong', 'R/W Type': 'PyTango.READ'}}
    MaxDevice = 22

    def __init__(self, inst, props):
        MotorController.MotorController.__init__(self, inst, props)
        print "PYTHON -> IcePapController ctor for instance", inst

        self.home_speed = 0.0
        self.nb_call = 0
        self.timeout = 3

        print "PYTHON -> IcePap on", self.Host, " and port", self.Port, " with timeout = ", self.timeout
        print "PYTHON -> Connected to", self.Host, " on port", self.Port

    def AddDevice(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In AddDevice method for axis", axis

    def DeleteDevice(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In DeleteDevice method for axis", axis

    def StateOne(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In StateOne method for axis", axis
        tup = (PyTango.DevState.ON, 0)
        return tup

    def PreReadAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreReadAll method"

    def PreReadOne(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreReadOne method for axis", axis

    def ReadAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In ReadAll method"

    def ReadOne(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In ReadOne method for axis", axis
        return 123

    def PreStartAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreStartAll method"

    def PreStartOne(self, axis, pos):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreStartOne method for axis", axis, " with pos", pos
        return True

    def StartOne(self, axis, pos):
        print "PYTHON -> IcePapController/", self.inst_name, ": In StartOne method for axis", axis, " with pos", pos

    def StartAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In StartAll method"

    def SetPar(self, axis, name, value):
        print "PYTHON -> IcePapController/", self.inst_name, ": In SetPar method for axis", axis, " name=", name, " value=", value

    def GetPar(self, axis, name):
        print "PYTHON -> IcePapController/", self.inst_name, ": In GetPar method for axis", axis, " name=", name
        return 12.34

    def GetExtraAttributePar(self, axis, name):
        print "PYTHON -> IcePapController/", self.inst_name, ": In GetExtraAttributePar method for axis", axis, " name=", name
        return 2233

    def SetExtraAttributePar(self, axis, name, value):
        print "PYTHON -> IcePapController/", self.inst_name, ": In SetExtraAttributePar method for axis", axis, " name=", name, " value=", value

    def __del__(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": Aarrrrrg, I am dying"

if __name__ == "__main__":
    obj = IcePapController('test')



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
channelWidgets.py:
"""

__all__ = ["PoolChannel", "PoolChannelTV", "_PoolChannelTV"]

import weakref

import tango
import taurus
from taurus.core import DataType, DataFormat
from taurus.external.qt import Qt
from taurus.qt.qtcore.mimetypes import (TAURUS_DEV_MIME_TYPE,
                                        TAURUS_ATTR_MIME_TYPE)
from taurus.qt.qtgui.panel import (TaurusValue, TaurusDevButton,
                                   DefaultLabelWidget, DefaultUnitsWidget,
                                   TaurusAttrForm, TaurusForm,
                                   DefaultReadWidgetLabel, TaurusPlotButton,
                                   TaurusImageButton, TaurusValuesTableButton,
                                   )
from taurus.qt.qtgui.input import TaurusValueLineEdit
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox
from taurus.qt.qtgui.compact import TaurusReadWriteSwitcher
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.resource import getIcon
from sardana.taurus.qt.qtgui.extra_pool.poolmotor import \
    LabelWidgetDragsDeviceAndAttribute, TaurusAttributeListener


class PoolChannelTVLabelWidget(TaurusWidget):
    """
    @TODO tooltip should be extended with status info
    @TODO context menu should be the lbl_alias extended
    @TODO default tooltip extended with the complete (multiline) status
    @TODO rightclick popup menu with actions: (1) switch user/expert view,
          (2) Config -all attributes-, (3) change channel
          For the (3), a drop event should accept if it is a device,
          and add it to the "change-channel" list and select
    """

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.setLayout(Qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.lbl_alias = DefaultLabelWidget(parent, designMode)
        self.lbl_alias.setBgRole("none")
        self.layout().addWidget(self.lbl_alias)

        # I don't like this approach, there should be something like
        # self.lbl_alias.addAction(...)
        self.lbl_alias.contextMenuEvent = \
            lambda event: self.contextMenuEvent(event)

        # I' don't like this approach, there should be something like
        # self.lbl_alias.addToolTipCallback(self.calculate_extra_tooltip)
        self.lbl_alias.getFormatedToolTip = self.calculateExtendedTooltip

        # I' don't like this approach, there should be something like
        # self.lbl_alias.disableDrag() or self.lbl_alias.setDragEnabled(False)
        # or better, define if Attribute or Device or Both have to be included
        # in the mimeData
        self.lbl_alias.mouseMoveEvent = self.mouseMoveEvent

    def setModel(self, model):
        if model in (None, ""):
            self.lbl_alias.setModel(model)
            TaurusWidget.setModel(self, model)
            return
        self.lbl_alias.taurusValueBuddy = self.taurusValueBuddy
        self.lbl_alias.setModel(model)
        TaurusWidget.setModel(self, model + "/Status")

    def calculateExtendedTooltip(self, cache=False):
        default_label_widget_tooltip = DefaultLabelWidget.getFormatedToolTip(
            self.lbl_alias, cache)
        status_info = ""
        channel_dev = self.taurusValueBuddy().channel_dev
        if channel_dev is not None:
            status = channel_dev.getAttribute("Status").read().value
            # MAKE IT LOOK LIKE THE STANDARD TABLE FOR TAURUS TOOLTIPS
            status_lines = status.split("\n")
            status_info = ("<TABLE width='500' border='0' cellpadding='1' "
                           "cellspacing='0'><TR><TD WIDTH='80' ALIGN='RIGHT'"
                           "VALIGN='MIDDLE'><B>Status:</B></TD><TD>"
                           + status_lines[0]
                           + "</TD></TR>")
            for status_extra_line in status_lines[1:]:
                status_info += ("<TR><TD></TD><TD>"
                                + status_extra_line
                                + "</TD></TR>")
            status_info += "</TABLE>"
        return default_label_widget_tooltip + status_info

    def contextMenuEvent(self, event):
        # Overwrite the default taurus label behaviour
        menu = Qt.QMenu(self)
        action_tango_attributes = Qt.QAction(self)
        action_tango_attributes.setIcon(
            getIcon(":/categories/preferences-system.svg"))
        action_tango_attributes.setText("Tango Attributes")
        menu.addAction(action_tango_attributes)
        action_tango_attributes.triggered.connect(
            self.taurusValueBuddy().showTangoAttributes)

        cm_action = menu.addAction("Compact")
        cm_action.setCheckable(True)
        cm_action.setChecked(self.taurusValueBuddy().isCompact())
        cm_action.toggled.connect(self.taurusValueBuddy().setCompact)

        menu.exec_(event.globalPos())
        event.accept()

    def mouseMoveEvent(self, event):
        model = self.taurusValueBuddy().getModelObj()
        mimeData = Qt.QMimeData()
        mimeData.setText(self.lbl_alias.text())
        dev_name = model.getFullName().encode("utf-8")
        attr_name = dev_name + b"/Value"
        mimeData.setData(TAURUS_DEV_MIME_TYPE, dev_name)
        mimeData.setData(TAURUS_ATTR_MIME_TYPE, attr_name)

        drag = Qt.QDrag(self)
        drag.setMimeData(mimeData)
        drag.setHotSpot(event.pos() - self.rect().topLeft())
        drag.exec_(Qt.Qt.CopyAction, Qt.Qt.CopyAction)


class _IntegTimeTaurusValueLineEdit(TaurusValueLineEdit):

    def sizeHint(self):
        size = Qt.QSize()
        size.setWidth(55)
        return size


class PoolChannelTVExtraWidget(TaurusWidget):
    """
    """

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.setLayout(Qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.le_integ_time = _IntegTimeTaurusValueLineEdit()
        le_policy = Qt.QSizePolicy(Qt.QSizePolicy.Preferred,
                                   Qt.QSizePolicy.Preferred)
        self.le_integ_time.setSizePolicy(le_policy)
        self.layout().addWidget(self.le_integ_time)

        # TODO: state listener should be part of *ExpChannel device Taurus
        #       Qt extension, it is necessary to change button's action,
        #       icon, etc.
        self.state_listener = TaurusAttributeListener()
        self.state_listener.eventReceivedSignal.connect(self.updateState)
        self.btn_start_stop_clicked_slot = None
        self.btn_start_stop = Qt.QPushButton()
        btn_policy = Qt.QSizePolicy(Qt.QSizePolicy.Fixed,
                                    Qt.QSizePolicy.Preferred)
        btn_policy.setHorizontalStretch(0)
        self.btn_start_stop.setSizePolicy(btn_policy)
        self.layout().addWidget(self.btn_start_stop)

    def updateState(self, state):
        btn_start_stop = self.btn_start_stop
        if self.btn_start_stop_clicked_slot is not None:
            btn_start_stop.clicked.disconnect(
                self.btn_start_stop_clicked_slot)
        if state == tango.DevState.MOVING:
            btn_start_stop.setToolTip("Stop the channel")
            btn_start_stop.setIcon(
                getIcon(":/actions/media_playback_stop.svg"))
            self.btn_start_stop_clicked_slot = self.abort
        else:
            btn_start_stop.setToolTip("Start the channel")
            btn_start_stop.setIcon(
                getIcon(":/actions/media_playback_start.svg"))
            self.btn_start_stop_clicked_slot = self.start
        btn_start_stop.clicked.connect(self.btn_start_stop_clicked_slot)

    def setModel(self, model):
        # first disconnect old model state listener
        model_obj = self.getModelObj()
        if model_obj is not None:
            model_obj.getAttribute("State").removeListener(
                self.state_listener)
        TaurusWidget.setModel(self, model)
        if model in (None, ""):
            self.le_integ_time.setModel(model)
            return
        self.le_integ_time.setModel(model + "/IntegrationTime")
        # connect new model state listener
        self.getModelObj().getAttribute("State").addListener(
            self.state_listener)

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(
        msg="An error occurred trying to start the acquisition.")
    def start(self):
        channel_dev = self.taurusValueBuddy().channel_dev
        if channel_dev is not None:
            channel_dev.Start()

    @Qt.pyqtSlot()
    @ProtectTaurusMessageBox(
        msg="An error occurred trying to abort the acquisition.")
    def abort(self):
        channel_dev = self.taurusValueBuddy().channel_dev
        if channel_dev is not None:
            channel_dev.abort()


class PoolChannelTVUnitsWidget(DefaultUnitsWidget):

    def __init__(self, parent=None, designMode=False):
        DefaultUnitsWidget.__init__(self, parent, designMode)

    def setModel(self, model):
        if model in (None, ""):
            DefaultUnitsWidget.setModel(self, model)
            return
        DefaultUnitsWidget.setModel(self, model + "/Value")


class PoolChannelTV(TaurusValue):
    """A widget that displays and controls a pool channel device.
    It differs from :class:`PoolChannel` in that it behaves as a TaurusValue
    (i.e., it allows its subwidgets to be aligned in columns in a TaurusForm)`

    .. todo:: Ideally overriding of `getDefaultReadWidgetClass` and
      `updateReadWidget` should not be necessary. Creation of a dedicated wiget
      for displaying value should be delegated to a custom read widget.
    .. todo:: draw state-based coloured frame around spectrum and image
       buttons, when state is MOVING - blue, when ON - green, etc.
    """

    def __init__(self, parent=None, designMode=False):
        TaurusValue.__init__(self, parent=parent, designMode=designMode)
        self.setLabelWidgetClass(PoolChannelTVLabelWidget)
        self.setWriteWidgetClass(None)
        self.setUnitsWidgetClass(PoolChannelTVUnitsWidget)
        self.setExtraWidgetClass(PoolChannelTVExtraWidget)
        self.channel_dev = None

    def getDefaultReadWidgetClass(self, returnAll=False):
        """
        Returns the default class (or classes) to use as read widget for the
        current model.

        Override TaurusValue.getDefaultReadWidgetClass. Simply do the same
        but based on the Value attribute while our model obj is a device.

        :param returnAll: (bool) if True, the return value is a list of valid
                          classes instead of just one class

        :return: (class or list<class>) the default class  to use for the read
                 widget (or, if returnAll==True, a list of classes that can
                 show the attribute ). If a list is returned, it will be
                 loosely ordered by preference, being the first element
                 always the default one.
        """
        modelobj = self.getModelObj()
        if modelobj is None:
            if returnAll:
                return [DefaultReadWidgetLabel]
            else:
                return DefaultReadWidgetLabel

        valueobj = modelobj.getAttribute("Value")
        if valueobj.data_format == DataFormat._0D:
            result = [DefaultReadWidgetLabel]
        elif valueobj.data_format == DataFormat._1D:
            if valueobj.type in (DataType.Float, DataType.Integer):
                result = [TaurusPlotButton,
                          TaurusValuesTableButton, DefaultReadWidgetLabel]
            else:
                result = [TaurusValuesTableButton, DefaultReadWidgetLabel]
        elif valueobj.data_format == DataFormat._2D:
            if valueobj.type in (DataType.Float, DataType.Integer):
                try:
                    # unused import but useful to determine if
                    # TaurusImageButton should be added
                    from taurus.qt.qtgui.extra_guiqwt import TaurusImageDialog  # noqa
                    result = [TaurusImageButton,
                              TaurusValuesTableButton, DefaultReadWidgetLabel]
                except ImportError:
                    result = [TaurusValuesTableButton,
                              DefaultReadWidgetLabel]
            else:
                result = [TaurusValuesTableButton, DefaultReadWidgetLabel]
        else:
            self.warning('Unsupported attribute type %s' % valueobj.type)
            result = None

        if returnAll:
            return result
        else:
            return result[0]

    def updateReadWidget(self):
        """Update read widget by recreating it from scratch.

        Override TaurusValue.updateReadWidget. Simply do the same, just
        don't call setModel on the read widget at the end. Model of the read
        widget is set by our setModel when the read widget is already
        recreated.
        """
        # get the class for the widget and replace it if necessary
        try:
            klass = self.readWidgetClassFactory(self.readWidgetClassID)
            self._readWidget = self._newSubwidget(self._readWidget, klass)
        except Exception as e:
            self._destroyWidget(self._readWidget)
            self._readWidget = Qt.QLabel('[Error]')
            msg = 'Error creating read widget:\n' + str(e)
            self._readWidget.setToolTip(msg)
            self.debug(msg)

        # take care of the layout
        self.addReadWidgetToLayout()

        if self._readWidget is not None:
            # give the new widget a reference to its buddy TaurusValue object
            self._readWidget.taurusValueBuddy = weakref.ref(self)
            if isinstance(self._readWidget, TaurusReadWriteSwitcher):
                self._readWidget.readWidget.taurusValueBuddy = weakref.ref(
                    self)
                self._readWidget.writeWidget.taurusValueBuddy = weakref.ref(
                    self)

            # tweak the new widget
            if self.minimumHeight() is not None:
                self._readWidget.setMinimumHeight(self.minimumHeight())

    def setModel(self, model):
        TaurusValue.setModel(self, model)
        if model == "" or model is None:
            return
        self.readWidget().setModel(model + "/Value")
        self.channel_dev = taurus.Device(model)

    def showTangoAttributes(self):
        model = self.getModel()
        taurus_attr_form = TaurusAttrForm()
        taurus_attr_form.setMinimumSize(Qt.QSize(555, 800))
        taurus_attr_form.setModel(model)
        taurus_attr_form.setWindowTitle(
            '%s Tango Attributes' % self.getModelObj().getSimpleName())
        taurus_attr_form.show()


class _ParentDevButton(TaurusDevButton):
    '''A TaurusDevButton that receives an attribute name but sets
    the corresponding device as model. **For internal use only** '''

    def __init__(self, **kwargs):
        TaurusDevButton.__init__(self, **kwargs)
        self.setText('')
        self.setSizePolicy(Qt.QSizePolicy.Preferred, Qt.QSizePolicy.Maximum)

    def setModel(self, model):
        try:
            attr = taurus.Attribute(model)
        except:
            return
        dev = attr.getParentObj()
        devname = dev.getFullName()
        TaurusDevButton.setModel(self, devname)


class _PoolChannelTV(TaurusValue):
    ''' A widget that displays and controls a pool channel device.
    It differs from :class:`PoolChannel` in that it behaves as a TaurusValue
    (i.e., it allows its subwidgets to be aligned in columns in a TaurusForm)`
    '''

    def __init__(self, parent=None, designMode=False):
        TaurusValue.__init__(self, parent=parent, designMode=designMode)
        self.setLabelWidgetClass(LabelWidgetDragsDeviceAndAttribute)
        self.setLabelConfig('{dev.name}')

    def getDefaultExtraWidgetClass(self):
        return _ParentDevButton

    def setModel(self, model):
        if model is not None:
            # @todo: change this (it assumes tango naming!)
            model = "%s/value" % model
        TaurusValue.setModel(self, model)

    def showEvent(self, event):
        TaurusValue.showEvent(self, event)
        try:
            self.getModelObj().getParentObj().getAttribute('Value').enablePolling(force=True)
        except:
            pass

    def hideEvent(self, event):
        TaurusValue.hideEvent(self, event)
        try:
            self.getModelObj().getParentObj().getAttribute('Value').disablePolling()
        except:
            pass



class PoolChannel(TaurusWidget):
    ''' A widget that displays and controls a pool channel device

    .. seealso:: :class:`PoolChannelTV`
    '''

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent)

        self.setLayout(Qt.QHBoxLayout())

        # put a widget with a TaurusValue
        w = Qt.QWidget()
        w.setLayout(Qt.QGridLayout())
        self._TaurusValue = TaurusValue(parent=w, designMode=designMode)
        self._TaurusValue.setLabelWidgetClass(
            LabelWidgetDragsDeviceAndAttribute)
        self._TaurusValue.setLabelConfig('{dev.name}')
        self.layout().addWidget(w)

        #...and a dev button next to the widget
        self._devButton = TaurusDevButton(parent=self, designMode=designMode)
        self._devButton.setText('')
        self.layout().addWidget(self._devButton)

        self.modelChanged.connect(self._updateTaurusValue)

    def _updateTaurusValue(self):
        m = self.getModelName()
        self._TaurusValue.setModel("%s/value" % m)
        self._devButton.setModel(m)


if __name__ == "__main__":
    import sys
    argv = sys.argv
    if len(argv) > 0:
        models = argv[1:]
    app = Qt.QApplication(sys.argv)

    form_tv = TaurusForm()
    form_tv.setModifiableByUser(True)
    tv_widget_class = "sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV"
    tv_class_map = {"CTExpChannel": (tv_widget_class, (), {}),
                    "OneDExpChannel": (tv_widget_class, (), {}),
                    "TwoDExpChannel": (tv_widget_class, (), {})}
    form_tv.setCustomWidgetMap(tv_class_map)
    form_tv.setModel(models)

    w = Qt.QWidget()
    w.setLayout(Qt.QVBoxLayout())
    w.layout().addWidget(form_tv)

    w.show()
    sys.exit(app.exec_())



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import time
import copy

from sardana import State
from sardana.pool import AcqSynch
from sardana.pool.controller import CounterTimerController, Type, Description


class Channel(object):

    def __init__(self, idx):
        self.idx = idx            # 1 based index
        self.value = 0.0
        self.is_counting = False
        self.acq_idx = 0
        self.buffer_values = []


class DummyCounterTimerController(CounterTimerController):
    """This class is the Tango Sardana CounterTimer controller for tests"""

    gender = "Simulation"
    model = "Basic"
    organization = "Sardana team"

    MaxDevice = 1024

    default_timer = 1
    default_latency_time = 0.0

    ctrl_attributes = {
        "Synchronizer": {
                Type: str,
                Description: ("Hardware (external) emulated synchronizer. "
                              "Can be any of dummy trigger/gate elements "
                              "from the same pool.")
            },
    }

    def __init__(self, inst, props, *args, **kwargs):
        CounterTimerController.__init__(self, inst, props, *args, **kwargs)
        self._synchronization = AcqSynch.SoftwareTrigger
        self.channels = self.MaxDevice * [None, ]
        self.start_time = None
        self.integ_time = None
        self.monitor_count = None
        self.repetitions = None
        self.latency_time = None
        self.acq_cycle_time = None
        self.estimated_duration = None
        self.read_channels = {}
        self.counting_channels = {}
        # name of synchronizer element
        self._synchronizer = None
        # synchronizer element (core)
        self.__synchronizer_obj = None
        # flag whether the controller was armed for hardware synchronization
        self._armed = False

    def AddDevice(self, axis):
        idx = axis - 1
        self.channels[idx] = Channel(axis)

    def DeleteDevice(self, axis):
        idx = axis - 1
        self.channels[idx] = None

    def LoadOne(self, axis, value, repetitions, latency_time):
        if value > 0:
            self.integ_time = integ_time = value
            self.monitor_count = None
            self.acq_cycle_time = acq_cycle_time = integ_time + latency_time
            self.estimated_duration =\
                acq_cycle_time * repetitions - latency_time
        else:
            self.integ_time = None
            self.monitor_count = -value

        self.repetitions = repetitions
        self.latency_time = latency_time

    def PreStartAll(self):
        self.counting_channels = {}

    def PreStartOne(self, axis, value=None):
        self._log.debug('PreStartOne(%d): entering...' % axis)
        idx = axis - 1
        channel = self.channels[idx]
        channel.value = 0.0
        channel.acq_idx = 0
        channel.buffer_values = []
        self.counting_channels[axis] = channel
        return True

    def StartOne(self, axis, value=None):
        self._log.debug('StartOne(%d): entering...' % axis)
        if self._synchronization in (AcqSynch.SoftwareStart,
                                     AcqSynch.SoftwareTrigger,
                                     AcqSynch.SoftwareGate):
            self.counting_channels[axis].is_counting = True

    def StartAll(self):
        if self._synchronization in (AcqSynch.HardwareStart,
                                     AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate):
            self._connect_hardware_synchronization()
            self._armed = True
        else:
            self.start_time = time.time()

    def StateOne(self, axis):
        self._log.debug('StateOne(%d): entering...' % axis)
        idx = axis - 1
        sta = State.On
        status = "Stopped"
        if self._armed:
            sta = State.Moving
            status = "Armed"
        elif axis in self.counting_channels and self.start_time is not None:
            channel = self.channels[idx]
            now = time.time()
            elapsed_time = now - self.start_time
            self._updateChannelState(axis, elapsed_time)
            if channel.is_counting:
                sta = State.Moving
                status = "Acquiring"
        ret = (sta, status)
        self._log.debug('StateOne(%d): returning %s' % (axis, repr(ret)))
        return sta, status

    def _updateChannelState(self, axis, elapsed_time):
        if self._synchronization == AcqSynch.SoftwareTrigger:
            if self.integ_time is not None:
                # counting in time
                if elapsed_time >= self.integ_time:
                    self._finish(elapsed_time)
            elif self.monitor_count is not None:
                # monitor counts
                v = int(elapsed_time * 100 * axis)
                if v >= self.monitor_count:
                    self._finish(elapsed_time)
        elif self._synchronization in (AcqSynch.HardwareTrigger,
                                       AcqSynch.HardwareGate,
                                       AcqSynch.SoftwareStart,
                                       AcqSynch.HardwareStart):
            if self.integ_time is not None:
                # counting in time
                if elapsed_time > self.estimated_duration:
                    self._finish(elapsed_time)

    def PreReadAll(self):
        self.read_channels = {}

    def PreReadOne(self, axis):
        channel = self.channels[axis - 1]
        self.read_channels[axis] = channel

    def ReadAll(self):
        if self._armed:
            return  # still armed - no trigger/gate arrived yet
        # if in acquisition then calculate the values to return
        if self.counting_channels and self.start_time is not None:
            now = time.time()
            elapsed_time = now - self.start_time
            for axis, channel in list(self.read_channels.items()):
                self._updateChannelState(axis, elapsed_time)
                if channel.is_counting:
                    self._updateChannelValue(axis, elapsed_time)

    def ReadOne(self, axis):
        self._log.debug('ReadOne(%d): entering...' % axis)
        channel = self.read_channels[axis]
        ret = None
        if self._synchronization in (AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate,
                                     AcqSynch.SoftwareStart,
                                     AcqSynch.HardwareStart):
            values = copy.deepcopy(channel.buffer_values)
            channel.buffer_values.__init__()
            ret = values
        elif self._synchronization == AcqSynch.SoftwareTrigger:
            ret = channel.value
        self._log.debug('ReadOne(%d): returning %s' % (axis, repr(ret)))
        return ret

    def _updateChannelValue(self, axis, elapsed_time):
        channel = self.channels[axis - 1]

        if self._synchronization == AcqSynch.SoftwareTrigger:
            if self.integ_time is not None:
                t = min([elapsed_time, self.integ_time])
                if axis == self._timer:
                    channel.value = t
                else:
                    channel.value = t * channel.idx
            elif self.monitor_count is not None:
                channel.value = int(elapsed_time * 100 * axis)
                if axis == self._monitor:
                    if not channel.is_counting:
                        channel.value = self.monitor_count
        elif self._synchronization in (AcqSynch.HardwareTrigger,
                                       AcqSynch.HardwareGate,
                                       AcqSynch.SoftwareStart,
                                       AcqSynch.HardwareStart):
            if self.monitor_count is not None:
                msg = ("count to monitor not supported in this "
                       "synchronization yet")
                raise NotImplementedError(msg)
            acq_cycle_time = self.acq_cycle_time
            nb_elapsed_acq, resting = divmod(elapsed_time, acq_cycle_time)
            nb_elapsed_acq = int(nb_elapsed_acq)
            # do not wait the last latency_time
            if (nb_elapsed_acq == self.repetitions - 1
                    and resting > self.integ_time):
                nb_elapsed_acq += 1
            if nb_elapsed_acq > self.repetitions:
                nb_elapsed_acq = self.repetitions
            nb_new_acq = nb_elapsed_acq - channel.acq_idx
            if nb_new_acq == 0:
                return
            if axis == self._timer:
                value = self.integ_time
            else:
                value = self.integ_time * channel.idx
            channel.buffer_values.extend([value] * nb_new_acq)
            channel.acq_idx = channel.acq_idx + nb_new_acq

    def _finish(self, elapsed_time, axis=None):
        if axis is None:
            for axis, channel in list(self.counting_channels.items()):
                channel.is_counting = False
                self._updateChannelValue(axis, elapsed_time)
        elif axis in self.counting_channels:
            channel = self.counting_channels[axis]
            channel.is_counting = False
            self._updateChannelValue(axis, elapsed_time)
            self.counting_channels.pop(axis)
        if self._synchronization in (AcqSynch.HardwareStart,
                                     AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate):
            self._disconnect_hardware_synchronization()
            self._armed = False
        self.start_time = None

    def AbortOne(self, axis):
        if axis not in self.counting_channels:
            return
        now = time.time()
        if self.start_time is not None:
            elapsed_time = now - self.start_time
        else:
            elapsed_time = 0
        self._finish(elapsed_time, axis)

    def GetCtrlPar(self, par):
        if par == 'synchronization':
            return self._synchronization
        elif par == 'latency_time':
            return self.default_latency_time

    def SetCtrlPar(self, par, value):
        if par == 'synchronization':
            self._synchronization = value

    def getSynchronizer(self):
        if self._synchronizer is None:
            return "None"
        else:
            # get synchronizer object to only check it exists
            self._synchronizer_obj
            return self._synchronizer

    def setSynchronizer(self, synchronizer):
        if synchronizer == "None":
            synchronizer = None
        self._synchronizer = synchronizer
        self.__synchronizer_obj = None  # invalidate cache

    @property
    def _synchronizer_obj(self):
        """Get synchronizer object with cache mechanism.

        If synchronizer object is not cached ("""
        if self.__synchronizer_obj is not None:
            return self.__synchronizer_obj
        synchronizer = self._synchronizer
        if synchronizer is None:
            msg = "Hardware (external) emulated synchronizer is not set"
            raise ValueError(msg)
        # getting pool (core) element - hack
        pool_ctrl = self._getPoolController()
        pool = pool_ctrl.pool
        try:
            synchronizer_obj = pool.get_element_by_name(synchronizer)
        except Exception:
            try:
                synchronizer_obj = pool.get_element_by_full_name(synchronizer)
            except Exception:
                msg = "Unknown synchronizer {0}".format(synchronizer)
                raise ValueError(msg)
        self.__synchronizer_obj = synchronizer_obj
        return synchronizer_obj

    def _connect_hardware_synchronization(self):
        # obtain dummy trigger/gate controller (plugin) instance - hack
        tg_ctrl = self._synchronizer_obj.controller.ctrl
        idx = self._synchronizer_obj.axis - 1
        func_generator = tg_ctrl.tg[idx]
        func_generator.add_listener(self)

    def _disconnect_hardware_synchronization(self):
        # obtain dummy trigger/gate controller (plugin) instance - hack
        tg_ctrl = self._synchronizer_obj.controller.ctrl
        idx = self._synchronizer_obj.axis - 1
        func_generator = tg_ctrl.tg[idx]
        func_generator.remove_listener(self)

    def event_received(self, src, type_, value):
        """Callback for dummy trigger/gate function generator events
        e.g. start, active passive
        """
        # for the moment only react on first trigger
        if type_.name.lower() == "start":
            self._armed = False
            for axis, channel in self.counting_channels.items():
                channel.is_counting = True
            self.start_time = time.time()



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""IORegister related macros"""

__docformat__ = 'restructuredtext'

__all__ = ["write_ioreg", "read_ioreg"]

from sardana.macroserver.macro import *

import array


class write_ioreg(Macro):
    """Writes a value to an input register"""

    param_def = [
        ['input/output register', Type.IORegister, None, 'input/output register'],
        ['data', Type.Float, None, 'value to be send']
    ]

    def run(self, ioreg, data):
        name = ioreg.getName()
        o = "Writing " + str(data) + " to " + name + " register "
        self.debug(o)
        data = ioreg.writeIORegister(data)


class read_ioreg(Macro):
    """Reads an output register"""

    param_def = [
        ['input/output register', Type.IORegister, None, 'input/output register']
    ]

    result_def = [
        ['input/output register value', Type.Float, None, 'value read from ' +
         'the input/output register']
    ]

    def run(self, ioreg):
        name = ioreg.getName()
        data = ioreg.readIORegister(force=True)
        data = float(data)
        o = "Reading " + name + " register "
        self.debug(o)
        return data



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import sys

import unittest
from taurus.test import insertTest


@insertTest(helper_name='prepare_waypoint', conf={"acc_time": 0.5,
                                                  "dec_time": 0.5,
                                                  "base_rate": 10,
                                                  "start_user_pos": 0,
                                                  "final_user_pos": 50},
            expected={"initial_pos": -60,
                      "final_pos": 135})
@insertTest(helper_name='prepare_waypoint', conf={"acc_time": 0.1,
                                                  "dec_time": 0.1,
                                                  "base_rate": 25,
                                                  "start_user_pos": 0,
                                                  "final_user_pos": 10},
            expected={"initial_pos": -1.25,
                      "final_pos": 16.25})
@insertTest(helper_name='prepare_waypoint', conf={"acc_time": 0.1,
                                                  "dec_time": 0.1,
                                                  "base_rate": 0,
                                                  "start_user_pos": 0,
                                                  "final_user_pos": 10},
            expected={"initial_pos": -2.5,
                      "final_pos": 17.5})
class CTScanTestCase(unittest.TestCase):

    def setUp(self):
        self.skipTest("CTScanTest is not ready to be run as part of testsuite")
#         modules = []
#         for mod in sys.modules.iterkeys():
#             if "sardana" in mod:
#                 modules.append(mod)
#         for mod in modules:
#             try:
#                 del sys.modules[mod]
#             except KeyError:
#                 pass

    @staticmethod
    def getEnv(name):
        if name == "ActiveMntGrp":
            return "MockMntGrp"
        elif name == "ScanID":
            return 1
        elif name == "ScanDir":
            return "/tmp"
        elif name == "ScanFile":
            return "MockFile.dat"
        from sardana.macroserver.msexception import UnknownEnv
        raise UnknownEnv

    def prepare_waypoint(self, conf, expected):
        try:
            from mock import MagicMock, patch
        except ImportError:
            self.skipTest("mock module is not available")
        with patch("sardana.macroserver.msparameter.Type"):
            from sardana.taurus.core.tango.sardana.pool import Motor
            from sardana.macroserver.scan.gscan import CTScan
            mock_motor = MagicMock(Motor)
            mock_motor.getBaseRate = MagicMock(return_value=conf["base_rate"])

            macro = MagicMock()
            macro.getMinAccTime
            macro.getEnv = self.getEnv
            macro.nr_interv = 2
            scan = CTScan(macro)
            scan.get_min_acc_time = MagicMock(return_value=conf["acc_time"])
            scan.get_min_dec_time = MagicMock(return_value=conf["dec_time"])
            scan._physical_moveables = [mock_motor]
            waypoint = {
                "positions": [conf["final_user_pos"]],
                "active_time": 0.3
            }
            start_positions = [conf["start_user_pos"]]
            ideal_paths, _, _ = scan.prepare_waypoint(
                waypoint, start_positions)
            path = ideal_paths[0]
            # Asserts
            msg = 'Initial positions do not match. (expected={0}, got={1})'.format(
                expected["initial_pos"], path.initial_pos)
            self.assertEqual(path.initial_pos, expected["initial_pos"], msg)
            msg = 'Final positions do not match. (expected={0}, got={1})'.format(
                expected["final_pos"], path.final_pos)
            self.assertEqual(path.final_pos, expected["final_pos"], msg)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

from sardana.pool.controller import PseudoCounterController, Type, MaxDimSize


class TwoDROI(PseudoCounterController):
    """A simple pseudo counter which receives an image from a 2D experimental
       channel and returns [1,1] quadrant"""

    counter_roles = "2D",
    pseudo_counter_roles = "Q1",

    def GetAxisAttributes(self, axis):
        axis_attrs = PseudoCounterController.GetAxisAttributes(self, axis)
        axis_attrs = dict(axis_attrs)
        axis_attrs['Value'][Type] = ((float,),)
        axis_attrs['Value'][MaxDimSize] = (256, 256)
        return axis_attrs

    def Calc(self, axis, counter_values):
        self._log.debug("Calc entering...")
        img = counter_values[0]
        if axis == 1:
            quadrant = img[0:255, 0:255]
        else:
            raise NotImplementedError("only first quadrant is implemented")
        self._log.debug("Calc returning: %r" % quadrant)
        return quadrant



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package provides the spock generic utilities"""


def translate_version_str2list(version_str):
    """Translates a version string in format 'x[.y[.z[...]]]' into a list of
    numbers"""
    if version_str is None:
        ver = [0, 0]
    else:
        ver = []
        for i in version_str.split(".")[:2]:
            try:
                i = int(i)
            except:
                i = 0
            ver.append(i)
    return ver


def get_ipython_version():
    """Returns the current IPython version"""
    import IPython
    v = None
    try:
        v = IPython.release.version
    except:
        pass
    return v


def get_ipython_version_list():
    ipv_str = get_ipython_version()
    return translate_version_str2list(ipv_str)


ipv = get_ipython_version_list()
if ipv < [1, 0]:
    raise Exception("IPython > 1 is required")
from .ipython_01_00.genutils import *  # noqa



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import unittest
from sardana.pool.poolcontrollermanager import ControllerManager


class ControllerManagerTestCase(unittest.TestCase):
    """Unittest of ControllerManager Class"""

    def setUp(self):
        """Create a ControllerManager"""
        self.cm = ControllerManager()

    def test_init(self):
        """Verify that the created ControllerManager is a ControllerManager
        instance."""
        msg = 'ControllerManager constructor does not create ' +\
              'ControllerManager instance'
        self.assertIsInstance(self.cm, ControllerManager, msg)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.cm = None



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import copy
import json
import os
import time
import threading

# TODO: decide what to use: taurus or PyTango
import PyTango
from PyTango import DeviceProxy
import unittest
from taurus.test import insertTest
from taurus.core.util import CodecFactory

from sardana import sardanacustomsettings
from sardana.pool import AcqSynchType, SynchDomain, SynchParam
from sardana.tango.pool.test import SarTestTestCase

from sardana.pool.test.util import AttributeListener


def _get_full_name(device_proxy, logger=None):
    """Obtain Sardana full name as it is used by the server."""
    db = PyTango.Database()
    if db.get_from_env_var():
        db_name = PyTango.ApiUtil.get_env_var("TANGO_HOST")
    else:
        host = device_proxy.get_db_host()  # this is FQDN
        port = device_proxy.get_db_port()
        db_name = host + ":" + port
    full_name = db_name + "/" + device_proxy.name()
    # try to use Taurus 4 to retrieve FQDN
    try:
        from taurus.core.tango.tangovalidator import TangoDeviceNameValidator
        full_name, _, _ = TangoDeviceNameValidator().getNames(full_name)
    # if Taurus3 in use just continue
    except ImportError:
        pass
    return full_name


class TangoAttributeListener(AttributeListener):

    def __init__(self, data_key="value"):
        AttributeListener.__init__(self)
        codec_name = getattr(sardanacustomsettings, "VALUE_BUFFER_CODEC")
        self._codec = CodecFactory().getCodec(codec_name)
        self._data_key = data_key

    def push_event(self, *args, **kwargs):
        self.event_received(*args, **kwargs)

    def event_received(self, *args, **kwargs):
        try:
            event = args[0]
            if event.err:
                for err in event.errors:
                    if err.reason == 'UnsupportedFeature':
                        # when subscribing for events, Tango does one
                        # readout of the attribute. However the Data
                        # attribute is not fereseen for readout, it is
                        # just the event communication channel.
                        # Ignoring this exception..
                        return
                    else:
                        raise err
            _, _value = self._codec.decode(event.attr_value.value)
            value = _value.get("value") or _value.get("value_ref")
            idx = _value['index']
            dev = event.device
            obj_fullname = _get_full_name(dev)
            # filling the measurement records
            with self.data_lock:
                channel_data = self.data.get(obj_fullname, [])
                expected_idx = len(channel_data)
                pad = [None] * (idx[0] - expected_idx)
                channel_data.extend(pad + value)
                self.data[obj_fullname] = channel_data
        except Exception as e:
            print(e)
            raise Exception('"data" event callback failed')


class MeasSarTestTestCase(SarTestTestCase):
    """ Helper class to setup the need environmet for execute """

    def setUp(self, pool_properties=None):
        SarTestTestCase.setUp(self, pool_properties)
        self.event_ids = {}
        self.mg_name = '_test_mg_1'

    def create_meas(self, config):
        """ Create a meas with the given configuration
        """
        # creating mg
        config = copy.deepcopy(config)
        self.expchan_names = []
        self.tg_names = []
        ordered_chns = [None] * 10
        for ctrl_name, ctrl_config in list(config.items()):
            channels = ctrl_config["channels"]
            for chn, chn_config in list(channels.items()):
                index = chn_config["index"]
                ordered_chns[index] = chn
        self.expchan_names = [chn for chn in ordered_chns if chn is not None]
        self.pool.CreateMeasurementGroup([self.mg_name] + self.expchan_names)

        for ctrl_name in list(config.keys()):
            ctrl_config = config.pop(ctrl_name)
            channels = ctrl_config["channels"]
            for chn_name in list(channels.keys()):
                chn_config = channels.pop(chn_name)
                chn_full_name = _get_full_name(DeviceProxy(chn_name))
                channels[chn_full_name] = chn_config
            ctrl_full_name = _get_full_name(DeviceProxy(ctrl_name))
            synchronizer = ctrl_config.get("synchronizer")
            if synchronizer is not None and synchronizer != "software":
                self.tg_names.append(synchronizer)
                synchronizer = _get_full_name(DeviceProxy(synchronizer))
                ctrl_config["synchronizer"] = synchronizer
            config[ctrl_full_name] = ctrl_config

        try:
            self.meas = PyTango.DeviceProxy(self.mg_name)
        except:
            raise Exception('Could not create the MeasurementGroup: %s' %
                            (self.mg_name))

        # When measurement group gets created it fills the configuration with
        # the default values. Reusing read configuration in order to set test
        # parameters.
        jcfg = self.meas.read_attribute('configuration').value
        cfg = json.loads(jcfg)
        for ctrl in cfg['controllers']:
            ctrl_config = cfg['controllers'][ctrl]
            ctrl_test_config = config[ctrl]
            channels = ctrl_config['channels']
            channels_test = ctrl_test_config.pop("channels")
            for chn, chn_config in list(channels.items()):
                chn_test_config = channels_test[chn]
                chn_config.update(chn_test_config)
            ctrl_config.update(ctrl_test_config)

        # Write the built configuration
        self.meas.write_attribute('configuration', json.dumps(cfg))

    def prepare_meas(self, params):
        """ Set the measurement group parameters
        """
        synch_description = params["synch_description"]
        codec = CodecFactory().getCodec('json')
        data = codec.encode(('', synch_description))
        self.meas.write_attribute('SynchDescription', data[1])

    def _add_attribute_listener(self, config):
        self.attr_listener = TangoAttributeListener()
        chn_names = []
        for ctrl_config in list(config.values()):
            for chn, chn_config in list(ctrl_config["channels"].items()):
                if chn_config.get("value_ref_enabled", False):
                    buffer_attr = "ValueRefBuffer"
                else:
                    buffer_attr = "ValueBuffer"
                dev = PyTango.DeviceProxy(chn)
                ch_fullname = _get_full_name(dev)
                event_id = dev.subscribe_event(buffer_attr,
                                               PyTango.EventType.CHANGE_EVENT,
                                               self.attr_listener)
                self.event_ids[dev] = event_id
                chn_names.append(ch_fullname)
        return chn_names

    def _acq_asserts(self, channel_names, repetitions):
        """ Do the asserts after an acquisition
        """
        # printing acquisition records
        table = self.attr_listener.get_table()
        header = table.dtype.names
        # checking if any of data was acquired
        self.assertTrue(self.attr_listener.data, 'no data were acquired')
        # checking if all channels produced data
        for channel in channel_names:
            msg = 'data from channel %s were not acquired' % channel
            self.assertIn(channel, header, msg)
        # checking if all the data were acquired
        for ch_name in header:
            ch_data_len = len(table[ch_name])
            msg = 'length of data for channel %s is %d and should be %d' %\
                (ch_name, ch_data_len, repetitions)
            self.assertEqual(ch_data_len, repetitions, msg)

    def meas_cont_acquisition(self, params, config):
        """ Helper method to do a continous acquisition
        """
        self.create_meas(config)
        self.prepare_meas(params)
        chn_names = self._add_attribute_listener(config)
        # Do acquisition
        self.meas.write_attribute("NbStarts", 1)
        self.meas.Prepare()
        self.meas.Start()
        while self.meas.State() == PyTango.DevState.MOVING:
            print("Acquiring...")
            time.sleep(0.1)
        time.sleep(1)
        repetitions = params['synch_description'][0][SynchParam.Repeats]
        self._acq_asserts(chn_names, repetitions)

    def push_event(self, event):
        value = event.attr_value.value
        self.meas_state = value
        if value == PyTango.DevState.MOVING:
            self.meas_started = True
        elif self.meas_started and value == PyTango.DevState.ON:
            self.meas_finished.set()

    def stop_meas_cont_acquisition(self, params, config):
        '''Helper method to do measurement and stop it'''
        self.create_meas(config)
        self.prepare_meas(params)
        self.meas_state = None
        self.meas_started = False
        self.meas_finished = threading.Event()
        chn_names = self._add_attribute_listener(config)
        # Do measurement
        id_ = self.meas.subscribe_event("State",
                                        PyTango.EventType.CHANGE_EVENT,
                                        self.push_event)
        try:
            # starting timer (0.2 s) which will stop the measurement group
            self.meas.write_attribute("NbStarts", 1)
            self.meas.Prepare()
            self.meas.Start()
            threading.Timer(0.2, self.stopMeas).start()
            self.assertTrue(self.meas_finished.wait(5), "mg has not stopped")
        finally:
            self.meas.unsubscribe_event(id_)
        desired_state = PyTango.DevState.ON
        msg = 'mg state after stop is %s (should be %s)' %\
            (self.meas_state, desired_state)
        self.assertEqual(self.meas_state, desired_state, msg)
        for name in chn_names:
            channel = PyTango.DeviceProxy(name)
            state = channel.state()
            msg = 'channel %s state after stop is %s (should be %s)' %\
                (name, state, desired_state)
            self.assertEqual(state, desired_state, msg)

    def stopMeas(self):
        '''Method used to stop measreument group'''
        self.meas.stop()

    def tearDown(self):
        for channel, event_id in list(self.event_ids.items()):
            channel.unsubscribe_event(event_id)
        try:
            # Delete the meas
            self.pool.DeleteElement(self.mg_name)
        except Exception as e:
            print('Impossible to delete MeasurementGroup: %s' %
                  self.mg_name)
            print(e)
        SarTestTestCase.tearDown(self)


synch_description1 = [{SynchParam.Delay: {SynchDomain.Time: 0},
                     SynchParam.Active: {SynchDomain.Time: .01},
                     SynchParam.Total: {SynchDomain.Time: .02},
                     SynchParam.Repeats: 10}]

params_1 = {
    "synch_description": synch_description1,
    "integ_time": 0.1,
    "name": '_exp_01'
}

synch_description2 = [{SynchParam.Delay: {SynchDomain.Time: 0},
                       SynchParam.Active: {SynchDomain.Time: 0.1},
                       SynchParam.Total: {SynchDomain.Time: 0.15},
                       SynchParam.Repeats: 10}]

params_2 = {
    "synch_description": synch_description2,
    "integ_time": 0.1,
    "name": '_exp_01'
}
doc_1 = 'Synchronized acquisition with two channels from the same controller'\
        ' using hardware trigger'
config_1 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
            "_test_ct_1_2": {
                "index": 1
            }
        }
    }
}

doc_2 = 'Synchronized acquisition with two channels from the same controller'\
        ' using software trigger'
config_2 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "software",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
            "_test_ct_1_2": {
                "index": 1
            }
        }
    }
}
doc_3 = 'Synchronized acquisition with four channels from two different'\
        'controllers using hardware and software triggers'
config_3 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "software",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
            "_test_ct_1_2": {
                "index": 1
            }
        }
    },
    "_test_ct_ctrl_2": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_2_1": {
                "index": 2
            },
            "_test_ct_2_2": {
                "index": 3
            }
        }
    },
}
doc_4 = 'Stop of the synchronized acquisition with two channels from the same'\
        ' controller using hardware trigger'
doc_5 = 'Stop of the synchronized acquisition with two channels from the same'\
        ' controller using software trigger'
doc_6 = 'Stop of the synchronized acquisition with four channels from two'\
        ' different controllers using hardware and software triggers'


@insertTest(helper_name='meas_cont_acquisition', test_method_doc=doc_1,
            params=params_1, config=config_1)
# TODO: implement dedicated asserts/test for only software synchronized
#       acquisition.
#       Until this TODO gets implemented we comment the test since it may
#       fail (last acquisitions may not be executed - the previous one could
#       still be in progress, so the shape of data will not correspond to the
#       number of repetitions
# @insertTest(helper_name='meas_cont_acquisition', test_method_doc=doc_2,
#             params=params_1, config=config_2)
@insertTest(helper_name='meas_cont_acquisition', test_method_doc=doc_3,
            params=params_1, config=config_3)
@insertTest(helper_name='stop_meas_cont_acquisition', test_method_doc=doc_4,
            params=params_2, config=config_1)
@insertTest(helper_name='stop_meas_cont_acquisition', test_method_doc=doc_5,
            params=params_2, config=config_2)
@insertTest(helper_name='stop_meas_cont_acquisition', test_method_doc=doc_6,
            params=params_2, config=config_3)
class TangoAcquisitionTestCase(MeasSarTestTestCase, unittest.TestCase):
    """Integration test of TGGeneration and Acquisition actions."""

    pass


config_5 = {
    "_test_2d_ctrl_1": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_2d_1_1": {
                "index": 1}
        },
    },
    "_test_roi_ctrl_1": {
        "channels": {
            "_test_roi_q1": {
                "index": 2
            }
        }
    }
}


@insertTest(helper_name='meas_cont_acquisition', test_method_doc="TODO",
            params=params_1, config=config_5)
class TangoAcquisition2DandPCTestCase(MeasSarTestTestCase, unittest.TestCase):

    pseudo_cls_list = (
        list(SarTestTestCase.pseudo_cls_list)
        + [("PseudoCounter", "ROI", "TwoDROI", "_test_roi", "1",
            "2D=_test_2d_1_1", "Q1=_test_roi_q1",)]
    )

    def setUp(self):
        ctrls_test_path = '../../../pool/test/res/controllers'
        source = os.path.join(os.path.dirname(__file__), ctrls_test_path)
        path = os.path.abspath(source)
        pool_properties = {'PoolPath': [path]}
        MeasSarTestTestCase.setUp(self, pool_properties)


config_6 = {
    "_test_2d_ctrl_1": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_2d_1_1": {
                "index": 1,
                "value_ref_enabled": True
            }
        },
    },
    "_test_roi_ctrl_1": {
        "channels": {
            "_test_roi_q1": {
                "index": 2
            }
        }
    }
}


@insertTest(helper_name='meas_cont_acquisition', test_method_doc="TODO",
            params=params_1, config=config_6)
class TangoAcquisition2DRefAndPCTestCase(MeasSarTestTestCase,
                                         unittest.TestCase):

    pseudo_cls_list = (
        list(SarTestTestCase.pseudo_cls_list)
        + [("PseudoCounter", "ROI", "TwoDROI", "_test_roi", "1",
            "2D=_test_2d_1_1", "Q1=_test_roi_q1",)]
    )

    def setUp(self):
        ctrls_test_path = '../../../pool/test/res/controllers'
        source = os.path.join(os.path.dirname(__file__), ctrls_test_path)
        path = os.path.abspath(source)
        pool_properties = {'PoolPath': [path]}
        MeasSarTestTestCase.setUp(self, pool_properties)
